"""Route executor for handling route lifecycle operations."""

import logging
from collections.abc import Mapping, Sequence
from typing import Any
from uuid import UUID

from ai.backend.common.clients.http_client.client_pool import ClientPool
from ai.backend.common.clients.valkey_client.valkey_schedule import (
    HealthCheckStatus,
    ValkeyScheduleClient,
)
from ai.backend.common.exception import BackendAIError
from ai.backend.common.service_discovery import ServiceDiscovery
from ai.backend.common.service_discovery.service_discovery import ModelServiceMetadata
from ai.backend.common.types import SessionId
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.data.deployment.types import DeploymentInfo, RouteStatus
from ai.backend.manager.errors.deployment import (
    EndpointNotFound,
    RouteSessionNotFound,
    RouteSessionTerminated,
    RouteUnhealthy,
)
from ai.backend.manager.repositories.deployment import DeploymentRepository
from ai.backend.manager.repositories.deployment.types import RouteData
from ai.backend.manager.repositories.scheduler.types.session_creation import SessionCreationSpec
from ai.backend.manager.sokovan.deployment.route.recorder.context import RouteRecorderContext
from ai.backend.manager.sokovan.deployment.route.types import (
    RouteExecutionError,
    RouteExecutionResult,
)
from ai.backend.manager.sokovan.scheduling_controller.scheduling_controller import (
    SchedulingController,
)

log = BraceStyleAdapter(logging.getLogger(__name__))


def _extract_error_code(exception: BaseException) -> str | None:
    """Extract error code from exception if available.

    Args:
        exception: The exception to extract error code from.

    Returns:
        Error code string if exception is BackendAIError, None otherwise.
    """
    if isinstance(exception, BackendAIError):
        return str(exception.error_code())
    return None


class RouteExecutor:
    """Executor for route lifecycle operations."""

    def __init__(
        self,
        deployment_repo: DeploymentRepository,
        scheduling_controller: SchedulingController,
        config_provider: ManagerConfigProvider,
        client_pool: ClientPool,
        valkey_schedule: ValkeyScheduleClient,
        service_discovery: ServiceDiscovery,
    ) -> None:
        self._deployment_repo = deployment_repo
        self._scheduling_controller = scheduling_controller
        self._config_provider = config_provider
        self._client_pool = client_pool
        self._valkey_schedule = valkey_schedule
        self._service_discovery = service_discovery

    async def provision_routes(self, routes: Sequence[RouteData]) -> RouteExecutionResult:
        """Provision routes by creating sessions.

        Args:
            routes: Routes to provision

        Returns:
            Result containing successful and failed routes
        """
        # Phase 1: Load configuration
        with RouteRecorderContext.shared_phase("load_configuration"):
            with RouteRecorderContext.shared_step("load_endpoint_config"):
                endpoint_ids = {route.endpoint_id for route in routes}
                deployments = await self._deployment_repo.get_endpoints_by_ids(endpoint_ids)
                deployment_map = {dep.id: dep for dep in deployments}

        route_session_ids: dict[UUID, SessionId] = {}
        successes: list[RouteData] = []
        errors: list[RouteExecutionError] = []

        # Phase 2: Create sessions (per-route)
        for route in routes:
            try:
                session_id = await self._provision_route(route, deployment_map)
                if session_id is not None:
                    route_session_ids[route.route_id] = session_id
                successes.append(route)
            except Exception as e:
                log.warning("Failed to provision route {}: {}", route.route_id, e)
                errors.append(
                    RouteExecutionError(
                        route_info=route,
                        reason="Failed to provision",
                        error_detail=str(e),
                        error_code=_extract_error_code(e),
                    )
                )

        # Phase 3: Link sessions to routes (only for successful routes)
        if route_session_ids:
            with RouteRecorderContext.shared_phase(
                "link_session_mapping", entity_ids=set(route_session_ids.keys())
            ):
                with RouteRecorderContext.shared_step("link_sessions_to_routes"):
                    await self._deployment_repo.update_route_sessions(route_session_ids)

        return RouteExecutionResult(
            successes=successes,
            errors=errors,
        )

    async def terminate_routes(self, routes: Sequence[RouteData]) -> RouteExecutionResult:
        """Terminate routes by destroying sessions.

        Args:
            routes: Routes to terminate

        Returns:
            Result containing successful and failed routes
        """
        # Collect session IDs from routes (no phase - cannot fail)
        target_session_ids: list[SessionId] = []
        for route in routes:
            if not route.session_id:
                log.debug("Route {} has no session, skipping termination", route.route_id)
                continue
            target_session_ids.append(route.session_id)

        # Phase 1: Terminate sessions
        with RouteRecorderContext.shared_phase("terminate_sessions"):
            with RouteRecorderContext.shared_step("mark_sessions_terminating"):
                await self._scheduling_controller.mark_sessions_for_termination(
                    target_session_ids, reason="ROUTE_TERMINATION"
                )

        return RouteExecutionResult(
            successes=list(routes),
            errors=[],
        )

    async def check_running_routes(self, routes: Sequence[RouteData]) -> RouteExecutionResult:
        """Check health status of running routes.

        Args:
            routes: Routes to check (should be HEALTHY or UNHEALTHY status)

        Returns:
            Result containing routes with updated health status
        """
        # Phase 1: Load status
        with RouteRecorderContext.shared_phase("load_status"):
            with RouteRecorderContext.shared_step("load_session_status"):
                route_ids = {route.route_id for route in routes}
                session_statuses = await self._deployment_repo.fetch_session_statuses_by_route_ids(
                    route_ids
                )

        successes: list[RouteData] = []
        errors: list[RouteExecutionError] = []

        # Phase 2: Verify status (per-route)
        for route in routes:
            try:
                self._verify_route_session_status(route, session_statuses)
                successes.append(route)
            except (RouteSessionNotFound, RouteSessionTerminated) as e:
                log.debug("Route {} session status check failed: {}", route.route_id, e)
                errors.append(
                    RouteExecutionError(
                        route_info=route,
                        reason=e.error_title,
                        error_detail=str(e),
                        error_code=_extract_error_code(e),
                    )
                )

        return RouteExecutionResult(
            successes=successes,
            errors=errors,
        )

    async def check_route_health(self, routes: Sequence[RouteData]) -> RouteExecutionResult:
        """
        Check health status of routes using Redis health data.

        Classifies routes into three states:
        - HEALTHY: readiness check passes and data is fresh (successes)
        - UNHEALTHY: readiness check fails (errors)
        - DEGRADED: health data is stale or missing (stale)

        Args:
            routes: Routes to check health for

        Returns:
            Result containing:
            - successes: healthy routes
            - errors: unhealthy routes
            - stale: degraded routes
        """
        # Phase 1: Load health status
        with RouteRecorderContext.shared_phase("load_health_status"):
            with RouteRecorderContext.shared_step("query_health_check_results"):
                # Get health status for all routes from Redis
                route_ids = [str(route.route_id) for route in routes]
                health_statuses = await self._valkey_schedule.check_route_health_status(route_ids)

        successes: list[RouteData] = []
        errors: list[RouteExecutionError] = []
        stale: list[RouteData] = []

        # Phase 2: Classify health state (per-route)
        for route in routes:
            try:
                is_healthy = self._classify_route_health(route, health_statuses)
                if is_healthy:
                    successes.append(route)
                else:
                    stale.append(route)
            except RouteUnhealthy as e:
                errors.append(
                    RouteExecutionError(
                        route_info=route,
                        reason=e.error_title,
                        error_detail=str(e),
                        error_code=_extract_error_code(e),
                    )
                )

        return RouteExecutionResult(
            successes=successes,
            errors=errors,
            stale=stale,
        )

    async def sync_service_discovery(self, routes: Sequence[RouteData]) -> RouteExecutionResult:
        """
        Sync healthy routes to service discovery backend for Prometheus scraping.

        For each route, extracts kernel host and port information and registers
        with service discovery so Prometheus can discover and scrape metrics endpoints.

        Args:
            routes: Healthy routes to sync

        Returns:
            Empty result (no status changes)
        """
        # Filter routes that have sessions (no phase - cannot fail)
        route_ids_with_session = {
            route.route_id for route in routes if route.session_id is not None
        }

        if not route_ids_with_session:
            log.debug("No routes with sessions to sync")
            return RouteExecutionResult(successes=[], errors=[])

        # Phase 1: Load service metadata
        with RouteRecorderContext.shared_phase("load_metadata"):
            with RouteRecorderContext.shared_step("load_service_metadata"):
                # Fetch service discovery information through repository
                route_discovery_data = (
                    await self._deployment_repo.fetch_route_service_discovery_info(
                        route_ids_with_session
                    )
                )

        # Construct ModelServiceMetadata for each route (no phase - transformation only)
        metadata_list: list[ModelServiceMetadata] = []
        for data in route_discovery_data:
            metadata = ModelServiceMetadata(
                route_id=data.route_id,
                model_service_name=data.endpoint_name,
                host=data.kernel_host,
                port=data.kernel_port,
                metrics_path="/metrics",
                labels={
                    "runtime_variant": data.runtime_variant,
                    "endpoint_id": str(data.endpoint_id),
                },
            )
            metadata_list.append(metadata)

        # Phase 2: Register for monitoring (only for routes with valid metadata)
        if metadata_list:
            with RouteRecorderContext.shared_phase(
                "register_monitoring", entity_ids={m.route_id for m in metadata_list}
            ):
                with RouteRecorderContext.shared_step("register_for_monitoring"):
                    await self._service_discovery.sync_model_service_routes(metadata_list)
            log.debug("Synced {} routes to service discovery", len(metadata_list))
        else:
            log.debug("No valid routes to sync to service discovery")

        return RouteExecutionResult(successes=[], errors=[])

    async def cleanup_routes_by_config(self, routes: Sequence[RouteData]) -> RouteExecutionResult:
        """
        Filter routes for cleanup based on scaling group configuration.

        Checks if each route's status (unhealthy/degraded) is in the scaling group's
        cleanup_target_statuses. Routes that should be cleaned up are returned as successes,
        others are filtered out.

        Args:
            routes: Routes to check for cleanup eligibility

        Returns:
            Result with routes that should be terminated (successes)
        """
        if not routes:
            return RouteExecutionResult(successes=[], errors=[])

        # Phase 1: Load cleanup configuration
        with RouteRecorderContext.shared_phase("load_cleanup_config"):
            with RouteRecorderContext.shared_step("load_endpoint_info"):
                endpoint_ids = {route.endpoint_id for route in routes}
                endpoints = await self._deployment_repo.get_endpoints_by_ids(endpoint_ids)

            with RouteRecorderContext.shared_step("load_cleanup_policy"):
                scaling_group_names = list({
                    endpoint.metadata.resource_group for endpoint in endpoints
                })
                cleanup_configs = await self._deployment_repo.get_scaling_group_cleanup_configs(
                    scaling_group_names
                )

        # Create mapping of endpoint_id -> cleanup config (no phase - transformation only)
        endpoint_cleanup_config: dict[UUID, set[RouteStatus]] = {}
        for endpoint in endpoints:
            config = cleanup_configs.get(endpoint.metadata.resource_group, None)
            if config:
                endpoint_cleanup_config[endpoint.id] = set(config.cleanup_target_statuses)
            else:
                endpoint_cleanup_config[endpoint.id] = set()

        successes: list[RouteData] = []

        # Phase 2: Identify cleanup targets (per-route)
        for route in routes:
            should_cleanup = self._check_route_cleanup_eligibility(route, endpoint_cleanup_config)
            if should_cleanup:
                successes.append(route)
                log.info(
                    "Route {} marked for cleanup (status: {})",
                    route.route_id,
                    route.status.value,
                )
            else:
                log.trace(
                    "Route {} kept (status {} not in cleanup targets)",
                    route.route_id,
                    route.status.value,
                )

        return RouteExecutionResult(
            successes=successes,
        )

    # Private helper methods

    async def _provision_route(
        self,
        route: RouteData,
        deployment_map: Mapping[UUID, DeploymentInfo],
    ) -> SessionId | None:
        """Provision a single route by creating a session.

        Returns:
            SessionId: newly created session ID
            None: route already has a session (skipped)
        """
        pool = RouteRecorderContext.current_pool()
        recorder = pool.recorder(route.route_id)

        with recorder.phase("create_session"):
            if route.session_id is not None:
                with recorder.step("skip_existing_session"):
                    log.debug("Route {} already has a session, skipping", route.route_id)
                return None

            with recorder.step("enqueue_session"):
                deployment = deployment_map.get(route.endpoint_id)
                if deployment is None:
                    raise EndpointNotFound(f"Deployment not found for endpoint {route.endpoint_id}")

                # Fetch deployment context with all necessary data
                deployment_context = await self._deployment_repo.fetch_deployment_context(
                    deployment
                )

                # Create session with full context
                return await self._scheduling_controller.enqueue_session(
                    SessionCreationSpec.from_deployment_info(
                        deployment_info=deployment,
                        context=deployment_context,
                        route_id=route.route_id,
                    )
                )

    def _verify_route_session_status(
        self,
        route: RouteData,
        session_statuses: Mapping[UUID, Any],
    ) -> None:
        """Verify that route's session is in a valid state."""
        pool = RouteRecorderContext.current_pool()
        recorder = pool.recorder(route.route_id)

        with recorder.phase("verify_session"):
            with recorder.step("check_session_exists"):
                session_status = session_statuses.get(route.route_id)
                if session_status is None:
                    raise RouteSessionNotFound("No session associated with route")

            with recorder.step("check_session_active"):
                if session_status.is_terminal():
                    raise RouteSessionTerminated(session_status.value)

    def _classify_route_health(
        self,
        route: RouteData,
        health_statuses: Mapping[str, Any],
    ) -> bool:
        """Classify route health status. Returns True for healthy, False for stale, raises for unhealthy."""
        pool = RouteRecorderContext.current_pool()
        recorder = pool.recorder(route.route_id)

        with recorder.phase("classify_health"):
            with recorder.step("determine_health_state"):
                route_id_str = str(route.route_id)
                health_status = health_statuses.get(route_id_str, None)

                if not health_status:
                    # No health data - Redis TTL expired, mark as stale
                    return False

                status = health_status.get_status()
                match status:
                    case HealthCheckStatus.HEALTHY:
                        return True
                    case HealthCheckStatus.STALE:
                        return False
                    case HealthCheckStatus.UNHEALTHY:
                        raise RouteUnhealthy("Route health check failed")

        return False  # Default to stale for unexpected cases

    def _check_route_cleanup_eligibility(
        self,
        route: RouteData,
        endpoint_cleanup_config: Mapping[UUID, set[RouteStatus]],
    ) -> bool:
        """Check if route should be cleaned up based on cleanup config."""
        pool = RouteRecorderContext.current_pool()
        recorder = pool.recorder(route.route_id)

        with recorder.phase("identify_cleanup_target"):
            with recorder.step("check_cleanup_eligibility"):
                cleanup_targets = endpoint_cleanup_config.get(route.endpoint_id, set())
                return route.status in cleanup_targets



"""Factory functions for creating post-processors."""

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from .base import KernelPostProcessor, PostProcessor
from .cache_invalidation import CacheInvalidationPostProcessor
from .kernel_schedule_marking import KernelScheduleMarkingPostProcessor
from .schedule_marking import ScheduleMarkingPostProcessor

if TYPE_CHECKING:
    from ai.backend.manager.repositories.scheduler.repository import SchedulerRepository
    from ai.backend.manager.sokovan.scheduling_controller import SchedulingController


def create_session_post_processors(
    scheduling_controller: SchedulingController,
    repository: SchedulerRepository,
) -> Sequence[PostProcessor]:
    """Create the default sequence of post-processors for session handlers.

    Args:
        scheduling_controller: Controller for marking schedules
        repository: Repository for cache invalidation

    Returns:
        Sequence of post-processors to execute in order
    """
    return [
        ScheduleMarkingPostProcessor(scheduling_controller),
        CacheInvalidationPostProcessor(repository),
    ]


def create_kernel_post_processors(
    scheduling_controller: SchedulingController,
) -> Sequence[KernelPostProcessor]:
    """Create the default sequence of post-processors for kernel handlers.

    Args:
        scheduling_controller: Controller for marking schedules

    Returns:
        Sequence of kernel post-processors to execute in order
    """
    return [
        KernelScheduleMarkingPostProcessor(scheduling_controller),
    ]



"""Processor package for export operations."""

from __future__ import annotations

from typing import override

from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor import ActionProcessor
from ai.backend.manager.actions.types import AbstractProcessorPackage, ActionSpec

from .actions import (
    ExportAuditLogsCSVAction,
    ExportAuditLogsCSVActionResult,
    ExportKeypairsCSVAction,
    ExportKeypairsCSVActionResult,
    ExportProjectsCSVAction,
    ExportProjectsCSVActionResult,
    ExportSessionsCSVAction,
    ExportSessionsCSVActionResult,
    ExportUsersCSVAction,
    ExportUsersCSVActionResult,
    GetReportAction,
    GetReportActionResult,
    ListReportsAction,
    ListReportsActionResult,
)
from .service import ExportService

__all__ = ("ExportProcessors",)


class ExportProcessors(AbstractProcessorPackage):
    """Processor package for export operations.

    Provides processors for:
    - Listing available reports
    - Getting report metadata
    - Report-specific CSV exports (users, sessions, projects, keypairs, audit-logs)
    """

    list_reports: ActionProcessor[ListReportsAction, ListReportsActionResult]
    get_report: ActionProcessor[GetReportAction, GetReportActionResult]
    export_users_csv: ActionProcessor[ExportUsersCSVAction, ExportUsersCSVActionResult]
    export_sessions_csv: ActionProcessor[ExportSessionsCSVAction, ExportSessionsCSVActionResult]
    export_projects_csv: ActionProcessor[ExportProjectsCSVAction, ExportProjectsCSVActionResult]
    export_keypairs_csv: ActionProcessor[ExportKeypairsCSVAction, ExportKeypairsCSVActionResult]
    export_audit_logs_csv: ActionProcessor[ExportAuditLogsCSVAction, ExportAuditLogsCSVActionResult]

    def __init__(self, service: ExportService, action_monitors: list[ActionMonitor]) -> None:
        self.list_reports = ActionProcessor(service.list_reports, action_monitors)
        self.get_report = ActionProcessor(service.get_report, action_monitors)
        self.export_users_csv = ActionProcessor(service.export_users_csv, action_monitors)
        self.export_sessions_csv = ActionProcessor(service.export_sessions_csv, action_monitors)
        self.export_projects_csv = ActionProcessor(service.export_projects_csv, action_monitors)
        self.export_keypairs_csv = ActionProcessor(service.export_keypairs_csv, action_monitors)
        self.export_audit_logs_csv = ActionProcessor(service.export_audit_logs_csv, action_monitors)

    @override
    def supported_actions(self) -> list[ActionSpec]:
        return [
            ListReportsAction.spec(),
            GetReportAction.spec(),
            ExportUsersCSVAction.spec(),
            ExportSessionsCSVAction.spec(),
            ExportProjectsCSVAction.spec(),
            ExportKeypairsCSVAction.spec(),
            ExportAuditLogsCSVAction.spec(),
        ]



"""Domain Fair Share query resolvers."""

from __future__ import annotations

import strawberry
from aiohttp import web
from strawberry import Info

from ai.backend.common.contexts.user import current_user
from ai.backend.manager.api.gql.fair_share.fetcher import (
    fetch_domain_fair_shares,
    fetch_rg_domain_fair_shares,
)
from ai.backend.manager.api.gql.fair_share.types import (
    BulkUpsertDomainFairShareWeightInput,
    BulkUpsertDomainFairShareWeightPayload,
    DomainFairShareConnection,
    DomainFairShareFilter,
    DomainFairShareGQL,
    DomainFairShareOrderBy,
    UpsertDomainFairShareWeightInput,
    UpsertDomainFairShareWeightPayload,
)
from ai.backend.manager.api.gql.types import ResourceGroupDomainScope, StrawberryGQLContext
from ai.backend.manager.api.gql.utils import check_admin_only
from ai.backend.manager.repositories.fair_share.types import (
    DomainFairShareSearchScope,
)
from ai.backend.manager.services.fair_share.actions import (
    BulkUpsertDomainFairShareWeightAction,
    DomainWeightInput,
    GetDomainFairShareAction,
    UpsertDomainFairShareWeightAction,
)

# Admin APIs


@strawberry.field(description="Added in 26.2.0. Get domain fair share data (admin only).")  # type: ignore[misc]
async def admin_domain_fair_share(
    info: Info[StrawberryGQLContext],
    resource_group_name: str,
    domain_name: str,
) -> DomainFairShareGQL | None:
    """Get a single domain fair share record (admin only)."""
    check_admin_only()

    processors = info.context.processors
    action_result = await processors.fair_share.get_domain_fair_share.wait_for_complete(
        GetDomainFairShareAction(
            resource_group=resource_group_name,
            domain_name=domain_name,
        )
    )

    return DomainFairShareGQL.from_dataclass(action_result.data)


@strawberry.field(description="Added in 26.2.0. List domain fair shares (admin only).")  # type: ignore[misc]
async def admin_domain_fair_shares(
    info: Info[StrawberryGQLContext],
    filter: DomainFairShareFilter | None = None,
    order_by: list[DomainFairShareOrderBy] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> DomainFairShareConnection | None:
    """Search domain fair shares with pagination (admin only)."""
    check_admin_only()

    return await fetch_domain_fair_shares(
        info=info,
        filter=filter,
        order_by=order_by,
        before=before,
        after=after,
        first=first,
        last=last,
        limit=limit,
        offset=offset,
    )


# Resource Group Scoped APIs


@strawberry.field(  # type: ignore[misc]
    description="Added in 26.2.0. Get domain fair share data within resource group scope."
)
async def rg_domain_fair_share(
    info: Info[StrawberryGQLContext],
    scope: ResourceGroupDomainScope,
    domain_name: str,
) -> DomainFairShareGQL | None:
    """Get a single domain fair share record within resource group scope."""
    processors = info.context.processors
    action_result = await processors.fair_share.get_domain_fair_share.wait_for_complete(
        GetDomainFairShareAction(
            resource_group=scope.resource_group_name,
            domain_name=domain_name,
        )
    )

    return DomainFairShareGQL.from_dataclass(action_result.data)


@strawberry.field(  # type: ignore[misc]
    description="Added in 26.2.0. List domain fair shares within resource group scope."
)
async def rg_domain_fair_shares(
    info: Info[StrawberryGQLContext],
    scope: ResourceGroupDomainScope,
    filter: DomainFairShareFilter | None = None,
    order_by: list[DomainFairShareOrderBy] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> DomainFairShareConnection | None:
    """Search domain fair shares within resource group scope."""
    repo_scope = DomainFairShareSearchScope(resource_group=scope.resource_group_name)
    return await fetch_rg_domain_fair_shares(
        info=info,
        scope=repo_scope,
        filter=filter,
        order_by=order_by,
        before=before,
        after=after,
        first=first,
        last=last,
        limit=limit,
        offset=offset,
    )


# Legacy APIs (deprecated)


@strawberry.field(  # type: ignore[misc]
    description="Added in 26.1.0. Get domain fair share data (superadmin only).",
    deprecation_reason=(
        "Use admin_domain_fair_share instead. "
        "This API will be removed after v26.3.0. See BEP-1041 for migration guide."
    ),
)
async def domain_fair_share(
    info: Info[StrawberryGQLContext],
    resource_group_name: str,
    domain_name: str,
) -> DomainFairShareGQL | None:
    """Get a single domain fair share record."""
    me = current_user()
    if me is None or not me.is_superadmin:
        raise web.HTTPForbidden(reason="Only superadmin can access fair share data.")

    processors = info.context.processors
    action_result = await processors.fair_share.get_domain_fair_share.wait_for_complete(
        GetDomainFairShareAction(
            resource_group=resource_group_name,
            domain_name=domain_name,
        )
    )

    return DomainFairShareGQL.from_dataclass(action_result.data)


@strawberry.field(  # type: ignore[misc]
    description="Added in 26.1.0. List domain fair shares (superadmin only).",
    deprecation_reason=(
        "Use admin_domain_fair_shares instead. "
        "This API will be removed after v26.3.0. See BEP-1041 for migration guide."
    ),
)
async def domain_fair_shares(
    info: Info[StrawberryGQLContext],
    filter: DomainFairShareFilter | None = None,
    order_by: list[DomainFairShareOrderBy] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> DomainFairShareConnection | None:
    """Search domain fair shares with pagination."""
    me = current_user()
    if me is None or not me.is_superadmin:
        raise web.HTTPForbidden(reason="Only superadmin can access fair share data.")

    return await fetch_domain_fair_shares(
        info=info,
        filter=filter,
        order_by=order_by,
        before=before,
        after=after,
        first=first,
        last=last,
        limit=limit,
        offset=offset,
    )


# Admin Mutations


@strawberry.mutation(  # type: ignore[misc]
    description=(
        "Added in 26.2.0. Upsert domain fair share weight (admin only). "
        "Creates a new record if it doesn't exist, or updates the weight if it does."
    )
)
async def admin_upsert_domain_fair_share_weight(
    info: Info[StrawberryGQLContext],
    input: UpsertDomainFairShareWeightInput,
) -> UpsertDomainFairShareWeightPayload:
    """Upsert domain fair share weight (admin only)."""
    check_admin_only()

    processors = info.context.processors
    action_result = await processors.fair_share.upsert_domain_fair_share_weight.wait_for_complete(
        UpsertDomainFairShareWeightAction(
            resource_group=input.resource_group_name,
            domain_name=input.domain_name,
            weight=input.weight,
        )
    )

    return UpsertDomainFairShareWeightPayload(
        domain_fair_share=DomainFairShareGQL.from_dataclass(action_result.data)
    )


@strawberry.mutation(  # type: ignore[misc]
    description=(
        "Added in 26.2.0. Bulk upsert domain fair share weights (admin only). "
        "Creates new records if they don't exist, or updates weights if they do."
    )
)
async def admin_bulk_upsert_domain_fair_share_weight(
    info: Info[StrawberryGQLContext],
    input: BulkUpsertDomainFairShareWeightInput,
) -> BulkUpsertDomainFairShareWeightPayload:
    """Bulk upsert domain fair share weights (admin only)."""
    check_admin_only()

    processors = info.context.processors
    action_result = (
        await processors.fair_share.bulk_upsert_domain_fair_share_weight.wait_for_complete(
            BulkUpsertDomainFairShareWeightAction(
                resource_group=input.resource_group_name,
                inputs=[
                    DomainWeightInput(
                        domain_name=item.domain_name,
                        weight=item.weight,
                    )
                    for item in input.inputs
                ],
            )
        )
    )

    return BulkUpsertDomainFairShareWeightPayload(upserted_count=action_result.upserted_count)


# Legacy Mutations (deprecated)


@strawberry.mutation(  # type: ignore[misc]
    description=(
        "Added in 26.1.0. Upsert domain fair share weight (superadmin only). "
        "Creates a new record if it doesn't exist, or updates the weight if it does."
    ),
    deprecation_reason=(
        "Use admin_upsert_domain_fair_share_weight instead. "
        "This API will be removed after v26.3.0. See BEP-1041 for migration guide."
    ),
)
async def upsert_domain_fair_share_weight(
    info: Info[StrawberryGQLContext],
    input: UpsertDomainFairShareWeightInput,
) -> UpsertDomainFairShareWeightPayload:
    """Upsert domain fair share weight."""
    me = current_user()
    if me is None or not me.is_superadmin:
        raise web.HTTPForbidden(reason="Only superadmin can modify fair share data.")

    processors = info.context.processors
    action_result = await processors.fair_share.upsert_domain_fair_share_weight.wait_for_complete(
        UpsertDomainFairShareWeightAction(
            resource_group=input.resource_group_name,
            domain_name=input.domain_name,
            weight=input.weight,
        )
    )

    return UpsertDomainFairShareWeightPayload(
        domain_fair_share=DomainFairShareGQL.from_dataclass(action_result.data)
    )


@strawberry.mutation(  # type: ignore[misc]
    description=(
        "Added in 26.1.0. Bulk upsert domain fair share weights (superadmin only). "
        "Creates new records if they don't exist, or updates weights if they do."
    ),
    deprecation_reason=(
        "Use admin_bulk_upsert_domain_fair_share_weight instead. "
        "This API will be removed after v26.3.0. See BEP-1041 for migration guide."
    ),
)
async def bulk_upsert_domain_fair_share_weight(
    info: Info[StrawberryGQLContext],
    input: BulkUpsertDomainFairShareWeightInput,
) -> BulkUpsertDomainFairShareWeightPayload:
    """Bulk upsert domain fair share weights."""
    me = current_user()
    if me is None or not me.is_superadmin:
        raise web.HTTPForbidden(reason="Only superadmin can modify fair share data.")

    processors = info.context.processors
    action_result = (
        await processors.fair_share.bulk_upsert_domain_fair_share_weight.wait_for_complete(
            BulkUpsertDomainFairShareWeightAction(
                resource_group=input.resource_group_name,
                inputs=[
                    DomainWeightInput(
                        domain_name=item.domain_name,
                        weight=item.weight,
                    )
                    for item in input.inputs
                ],
            )
        )
    )

    return BulkUpsertDomainFairShareWeightPayload(upserted_count=action_result.upserted_count)



"""GraphQL types and operations for app configuration."""

from __future__ import annotations

import strawberry
from strawberry import ID, Info

from ai.backend.common.contexts.user import current_user
from ai.backend.manager.api.gql.utils import check_admin_only, dedent_strip
from ai.backend.manager.errors.auth import InsufficientPrivilege
from ai.backend.manager.repositories.app_config.updaters import AppConfigUpdaterSpec
from ai.backend.manager.services.app_config.actions import (
    DeleteDomainConfigAction,
    DeleteUserConfigAction,
    GetDomainConfigAction,
    GetMergedAppConfigAction,
    GetUserConfigAction,
    UpsertDomainConfigAction,
    UpsertUserConfigAction,
)
from ai.backend.manager.types import OptionalState

from .types import StrawberryGQLContext


@strawberry.type(description="Added in 25.16.0. App configuration data")
class AppConfig:
    """GraphQL type for app configuration."""

    extra_config: strawberry.scalars.JSON


@strawberry.input(
    description=dedent_strip(
        """\
        Added in 25.16.0.
        Input for creating or updating domain-level app configuration.
        The provided extra_config object will completely replace the existing configuration;
        existing keys not present in the new extra_config will be removed.
        All users in this domain will be affected by these settings when their configurations are merged.
        """
    )
)
class UpsertDomainConfigInput:
    """Input type for upserting domain-level app configuration."""

    domain_name: str
    extra_config: strawberry.scalars.JSON

    def to_updater_spec(self) -> AppConfigUpdaterSpec:
        return AppConfigUpdaterSpec(extra_config=OptionalState.update(self.extra_config))


@strawberry.input(
    description=dedent_strip(
        """\
        Added in 25.16.0.
        Input for creating or updating user-level app configuration.
        The provided extra_config object will completely replace the existing configuration;
        existing keys not present in the new extra_config will be removed.
        These settings will override domain-level settings when configurations are merged for this user.
        If user_id is not provided, the current user's configuration will be updated.
        """
    )
)
class UpsertUserConfigInput:
    """Input type for upserting user-level app configuration."""

    extra_config: strawberry.scalars.JSON
    user_id: ID | None = None

    def to_updater_spec(self) -> AppConfigUpdaterSpec:
        return AppConfigUpdaterSpec(extra_config=OptionalState.update(self.extra_config))


@strawberry.input(description="Added in 25.16.0. Input for deleting domain-level app configuration")
class DeleteDomainConfigInput:
    """Input type for deleting domain-level app configuration."""

    domain_name: str


@strawberry.input(
    description=dedent_strip(
        """\
        Added in 25.16.0.
        Input for deleting user-level app configuration.
        If user_id is not provided, the current user's configuration will be deleted.
        """
    )
)
class DeleteUserConfigInput:
    """Input type for deleting user-level app configuration."""

    user_id: ID | None = None


@strawberry.type(
    description=dedent_strip(
        """\
        Added in 25.16.0.
        Payload returned after upserting domain-level app configuration.
        Contains the resulting configuration that was stored.
        """
    )
)
class UpsertDomainConfigPayload:
    """Payload returned after upserting domain-level app configuration."""

    app_config: AppConfig


@strawberry.type(
    description=dedent_strip(
        """\
        Added in 25.16.0.
        Payload returned after upserting user-level app configuration.
        Contains the resulting configuration that was stored.
        """
    )
)
class UpsertUserConfigPayload:
    """Payload returned after upserting user-level app configuration."""

    app_config: AppConfig


@strawberry.type(
    description=dedent_strip(
        """\
        Added in 25.16.0.
        Payload returned after deleting domain-level app configuration.
        Indicates whether the deletion was successful.
        """
    )
)
class DeleteDomainConfigPayload:
    """Payload returned after deleting domain-level app configuration."""

    deleted: bool


@strawberry.type(
    description=dedent_strip(
        """\
        Added in 25.16.0.
        Payload returned after deleting user-level app configuration.
        Indicates whether the deletion was successful.
        """
    )
)
class DeleteUserConfigPayload:
    """Payload returned after deleting user-level app configuration."""

    deleted: bool


@strawberry.field(  # type: ignore[misc]
    description=dedent_strip(
        """\
        Added in 26.2.0.
        Retrieve domain-level app configuration (admin only).
        Returns only the configuration set specifically for the domain, without merging.
        This query is useful for checking what values are configured at the domain level
        when you want to modify domain or user configurations separately.
        For actual configuration values to be applied, use mergedAppConfig instead.
        """
    )
)
async def admin_domain_app_config(
    domain_name: str,
    info: Info[StrawberryGQLContext],
) -> AppConfig | None:
    """Get domain-level app configuration (admin only)."""
    check_admin_only()
    processors = info.context.processors

    action_result = await processors.app_config.get_domain_config.wait_for_complete(
        GetDomainConfigAction(domain_name=domain_name)
    )

    if not action_result.result:
        return None

    return AppConfig(extra_config=action_result.result.extra_config)


@strawberry.field(  # type: ignore[misc]
    description=dedent_strip(
        """\
        Added in 25.16.0.
        Retrieve domain-level app configuration.
        Returns only the configuration set specifically for the domain, without merging.
        This query is useful for checking what values are configured at the domain level
        when you want to modify domain or user configurations separately.
        For actual configuration values to be applied, use mergedAppConfig instead.
        Requires admin privileges.
        """
    ),
    deprecation_reason=(
        "Use admin_domain_app_config instead. "
        "This API will be removed after v26.3.0. See BEP-1041 for migration guide."
    ),
)
async def domain_app_config(
    domain_name: str,
    info: Info[StrawberryGQLContext],
) -> AppConfig | None:
    """Get domain-level app configuration."""
    processors = info.context.processors
    me = current_user()
    if me is None or not (me.is_admin or me.is_superadmin):
        raise InsufficientPrivilege("Admin privileges required to access domain configuration")

    action_result = await processors.app_config.get_domain_config.wait_for_complete(
        GetDomainConfigAction(domain_name=domain_name)
    )

    if not action_result.result:
        return None

    return AppConfig(extra_config=action_result.result.extra_config)


@strawberry.field(  # type: ignore[misc]
    description=dedent_strip(
        """\
        Added in 25.16.0.
        Retrieve user-level app configuration.
        Returns only the configuration set specifically for the user, without merging with domain config.
        This query is useful for checking what values are configured at the user level
        when you want to modify domain or user configurations separately.
        For actual configuration values to be applied, use mergedAppConfig instead.
        If user_id is not provided, returns the current user's configuration.
        Users can only access their own configuration, but admins can access any user's configuration.
        """
    )
)
async def user_app_config(
    info: Info[StrawberryGQLContext],
    user_id: ID | None = None,
) -> AppConfig | None:
    """Get user-level app configuration."""
    processors = info.context.processors
    me = current_user()
    if me is None:
        raise InsufficientPrivilege("Authentication required")

    # Use current user's ID if user_id is not provided
    target_user_id = str(user_id) if user_id is not None else str(me.user_id)

    if str(me.user_id) != target_user_id and not (me.is_admin or me.is_superadmin):
        raise InsufficientPrivilege("Cannot access another user's app configuration")

    action_result = await processors.app_config.get_user_config.wait_for_complete(
        GetUserConfigAction(user_id=target_user_id)
    )

    if not action_result.result:
        return None

    return AppConfig(extra_config=action_result.result.extra_config)


@strawberry.field(  # type: ignore[misc]
    description=dedent_strip(
        """\
        Added in 25.16.0.
        Retrieve merged app configuration for the current user.
        The result combines domain-level and user-level configurations,
        where user settings override domain settings for the same keys.
        This query should be used when working with user app configurations
        to get the actual configuration values that will be applied.
        """
    )
)
async def merged_app_config(
    info: Info[StrawberryGQLContext],
) -> AppConfig:
    """Get merged app configuration for the current user."""
    processors = info.context.processors
    me = current_user()
    if me is None:
        raise InsufficientPrivilege("Authentication required")

    action_result = await processors.app_config.get_merged_config.wait_for_complete(
        GetMergedAppConfigAction(user_id=str(me.user_id))
    )

    return AppConfig(extra_config=action_result.merged_config)


@strawberry.mutation(  # type: ignore[misc]
    name="adminUpsertDomainAppConfig",
    description=dedent_strip(
        """\
        Added in 26.2.0.
        Create or update domain-level app configuration (admin only).
        The provided extra_config object will completely replace the existing configuration;
        existing keys not present in the new extra_config will be removed.
        All users in this domain will be affected by these settings when their configurations are merged,
        unless they have user-level configurations that override specific keys.
        """
    ),
)
async def admin_upsert_domain_app_config(
    input: UpsertDomainConfigInput,
    info: Info[StrawberryGQLContext],
) -> UpsertDomainConfigPayload:
    """Create or update domain-level app configuration (admin only)."""
    check_admin_only()
    processors = info.context.processors

    action_result = await processors.app_config.upsert_domain_config.wait_for_complete(
        UpsertDomainConfigAction(
            domain_name=input.domain_name,
            updater_spec=input.to_updater_spec(),
        )
    )

    return UpsertDomainConfigPayload(
        app_config=AppConfig(extra_config=action_result.result.extra_config)
    )


@strawberry.mutation(  # type: ignore[misc]
    name="upsertDomainAppConfig",
    description=dedent_strip(
        """\
        Added in 25.16.0.
        Create or update domain-level app configuration.
        The provided extra_config object will completely replace the existing configuration;
        existing keys not present in the new extra_config will be removed.
        All users in this domain will be affected by these settings when their configurations are merged,
        unless they have user-level configurations that override specific keys.
        Requires admin privileges.
        """
    ),
    deprecation_reason=(
        "Use admin_upsert_domain_app_config instead. "
        "This API will be removed after v26.3.0. See BEP-1041 for migration guide."
    ),
)
async def upsert_domain_app_config(
    input: UpsertDomainConfigInput,
    info: Info[StrawberryGQLContext],
) -> UpsertDomainConfigPayload:
    """Create or update domain-level app configuration."""
    processors = info.context.processors
    me = current_user()
    if me is None or not (me.is_admin or me.is_superadmin):
        raise InsufficientPrivilege("Admin privileges required to modify domain configuration")

    action_result = await processors.app_config.upsert_domain_config.wait_for_complete(
        UpsertDomainConfigAction(
            domain_name=input.domain_name,
            updater_spec=input.to_updater_spec(),
        )
    )

    return UpsertDomainConfigPayload(
        app_config=AppConfig(extra_config=action_result.result.extra_config)
    )


@strawberry.mutation(  # type: ignore[misc]
    name="upsertUserAppConfig",
    description=dedent_strip(
        """\
        Added in 25.16.0.
        Create or update user-level app configuration.
        The provided extra_config object will completely replace the existing configuration;
        existing keys not present in the new extra_config will be removed.
        These settings will override domain-level settings when configurations are merged for this user.
        If user_id is not provided, the current user's configuration will be updated.
        Users can only modify their own configuration, but admins can modify any user's configuration.
        """
    ),
)
async def upsert_user_app_config(
    input: UpsertUserConfigInput,
    info: Info[StrawberryGQLContext],
) -> UpsertUserConfigPayload:
    """Create or update user-level app configuration."""
    processors = info.context.processors
    me = current_user()
    if me is None:
        raise InsufficientPrivilege("Authentication required")

    # Use current user's ID if user_id is not provided
    target_user_id = str(input.user_id) if input.user_id is not None else str(me.user_id)

    if str(me.user_id) != target_user_id and not (me.is_admin or me.is_superadmin):
        raise InsufficientPrivilege("Cannot modify another user's app configuration")

    action_result = await processors.app_config.upsert_user_config.wait_for_complete(
        UpsertUserConfigAction(
            user_id=target_user_id,
            updater_spec=input.to_updater_spec(),
        )
    )

    return UpsertUserConfigPayload(
        app_config=AppConfig(extra_config=action_result.result.extra_config)
    )


@strawberry.mutation(  # type: ignore[misc]
    name="adminDeleteDomainAppConfig",
    description=dedent_strip(
        """\
        Added in 26.2.0.
        Delete domain-level app configuration (admin only).
        All users in this domain may be affected by this deletion.
        After deletion, users will only receive their user-level configurations
        when configurations are merged, with no domain-level defaults.
        """
    ),
)
async def admin_delete_domain_app_config(
    input: DeleteDomainConfigInput,
    info: Info[StrawberryGQLContext],
) -> DeleteDomainConfigPayload:
    """Delete domain-level app configuration (admin only)."""
    check_admin_only()
    processors = info.context.processors

    action_result = await processors.app_config.delete_domain_config.wait_for_complete(
        DeleteDomainConfigAction(domain_name=input.domain_name)
    )

    return DeleteDomainConfigPayload(deleted=action_result.deleted)


@strawberry.mutation(  # type: ignore[misc]
    name="deleteDomainAppConfig",
    description=dedent_strip(
        """\
        Added in 25.16.0.
        Delete domain-level app configuration.
        All users in this domain may be affected by this deletion.
        After deletion, users will only receive their user-level configurations
        when configurations are merged, with no domain-level defaults.
        Requires admin privileges.
        """
    ),
    deprecation_reason=(
        "Use admin_delete_domain_app_config instead. "
        "This API will be removed after v26.3.0. See BEP-1041 for migration guide."
    ),
)
async def delete_domain_app_config(
    input: DeleteDomainConfigInput,
    info: Info[StrawberryGQLContext],
) -> DeleteDomainConfigPayload:
    """Delete domain-level app configuration."""
    processors = info.context.processors
    me = current_user()
    if me is None or not (me.is_admin or me.is_superadmin):
        raise InsufficientPrivilege("Admin privileges required to delete domain configuration")

    action_result = await processors.app_config.delete_domain_config.wait_for_complete(
        DeleteDomainConfigAction(domain_name=input.domain_name)
    )

    return DeleteDomainConfigPayload(deleted=action_result.deleted)


@strawberry.mutation(  # type: ignore[misc]
    name="deleteUserAppConfig",
    description=dedent_strip(
        """\
        Added in 25.16.0.
        Delete user-level app configuration.
        After deletion, the user will still receive domain-level configuration values
        when configurations are merged, as domain settings remain unaffected.
        If user_id is not provided, the current user's configuration will be deleted.
        Users can only delete their own configuration, but admins can delete any user's configuration.
        """
    ),
)
async def delete_user_app_config(
    input: DeleteUserConfigInput,
    info: Info[StrawberryGQLContext],
) -> DeleteUserConfigPayload:
    """Delete user-level app configuration."""
    processors = info.context.processors
    me = current_user()
    if me is None:
        raise InsufficientPrivilege("Authentication required")

    # Use current user's ID if user_id is not provided
    target_user_id = str(input.user_id) if input.user_id is not None else str(me.user_id)

    if str(me.user_id) != target_user_id and not (me.is_admin or me.is_superadmin):
        raise InsufficientPrivilege("Cannot delete another user's app configuration")

    action_result = await processors.app_config.delete_user_config.wait_for_complete(
        DeleteUserConfigAction(user_id=target_user_id)
    )

    return DeleteUserConfigPayload(deleted=action_result.deleted)



"""CreatorSpec for deployment revision creation."""

from __future__ import annotations

import uuid
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from typing import Any, override

from ai.backend.common.types import (
    ResourceSlot,
    RuntimeVariant,
    VFolderMount,
)
from ai.backend.manager.models.deployment_revision import DeploymentRevisionRow
from ai.backend.manager.repositories.base import CreatorSpec


@dataclass
class DeploymentRevisionCreatorSpec(CreatorSpec[DeploymentRevisionRow]):
    """CreatorSpec for deployment revision creation.

    Note: revision_number must be provided by the service layer after
    calculating from get_latest_revision_number().
    """

    endpoint_id: uuid.UUID
    revision_number: int
    image_id: uuid.UUID
    resource_group: str
    resource_slots: ResourceSlot
    resource_opts: Mapping[str, Any]
    cluster_mode: str
    cluster_size: int
    model_id: uuid.UUID | None
    model_mount_destination: str
    model_definition_path: str | None
    model_definition: Mapping[str, Any] | None
    startup_command: str | None
    bootstrap_script: str | None
    environ: Mapping[str, Any]
    callback_url: str | None
    runtime_variant: RuntimeVariant
    extra_mounts: Sequence[VFolderMount]

    @override
    def build_row(self) -> DeploymentRevisionRow:
        return DeploymentRevisionRow(
            endpoint=self.endpoint_id,
            revision_number=self.revision_number,
            image=self.image_id,
            model=self.model_id,
            model_mount_destination=self.model_mount_destination,
            model_definition_path=self.model_definition_path,
            model_definition=self.model_definition,
            resource_group=self.resource_group,
            resource_slots=self.resource_slots,
            resource_opts=self.resource_opts,
            cluster_mode=self.cluster_mode,
            cluster_size=self.cluster_size,
            startup_command=self.startup_command,
            bootstrap_script=self.bootstrap_script,
            environ=self.environ,
            callback_url=self.callback_url,
            runtime_variant=self.runtime_variant,
            extra_mounts=list(self.extra_mounts),
        )



from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.artifact.types import ArtifactData
from ai.backend.manager.models.artifact import ArtifactRow
from ai.backend.manager.repositories.base.updater import Updater
from ai.backend.manager.services.artifact.actions.base import ArtifactAction


@dataclass
class UpdateArtifactAction(ArtifactAction):
    updater: Updater[ArtifactRow]

    @override
    def entity_id(self) -> str | None:
        return str(self.updater.pk_value)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE


@dataclass
class UpdateArtifactActionResult(BaseActionResult):
    result: ArtifactData

    @override
    def entity_id(self) -> str | None:
        return str(self.result.id)



from typing import override

from ai.backend.common.config import (
    ModelConfig,
    ModelDefinition,
    ModelHealthCheck,
    ModelServiceConfig,
)
from ai.backend.common.types import MODEL_SERVICE_RUNTIME_PROFILES, RuntimeVariant
from ai.backend.manager.data.deployment.types import (
    ModelRevisionSpec,
)
from ai.backend.manager.sokovan.deployment.definition_generator.base import ModelDefinitionGenerator


class ModularMAXModelDefinitionGenerator(ModelDefinitionGenerator):
    """Model definition generator implementation for Modular MAX runtime variant."""

    @override
    async def generate_model_definition(self, model_revision: ModelRevisionSpec) -> ModelDefinition:
        runtime_profile = MODEL_SERVICE_RUNTIME_PROFILES[RuntimeVariant.MODULAR_MAX]

        health_check: ModelHealthCheck | None = None
        if runtime_profile.health_check_endpoint:
            health_check = ModelHealthCheck(
                path=runtime_profile.health_check_endpoint,
                interval=10.0,
                max_retries=10,
                initial_delay=300.0,
            )

        model = ModelConfig(
            name="max-model",
            model_path=model_revision.mounts.model_mount_destination,
            service=ModelServiceConfig(
                start_command=model_revision.execution.startup_command or "",
                port=runtime_profile.port or 8000,
                health_check=health_check,
            ),
        )

        return ModelDefinition(models=[model])



"""Upserter specs for Fair Share repository upsert (INSERT ON CONFLICT UPDATE) operations."""

from __future__ import annotations

import uuid
from dataclasses import dataclass, field
from datetime import date
from decimal import Decimal
from typing import Any, override

from ai.backend.common.types import ResourceSlot
from ai.backend.manager.models.fair_share import (
    DomainFairShareRow,
    ProjectFairShareRow,
    UserFairShareRow,
)
from ai.backend.manager.repositories.base import UpserterSpec
from ai.backend.manager.types import OptionalState, TriState


@dataclass
class DomainFairShareUpserterSpec(UpserterSpec[DomainFairShareRow]):
    """Upserter spec for DomainFairShareRow.

    Unique constraint: (resource_group, domain_name)

    - Identity fields are required (for unique constraint)
    - Spec fields use Row defaults on INSERT if NOP
    - Calculation fields are only updated if UPDATE
    """

    # Identity (required)
    resource_group: str
    domain_name: str

    # Spec (uses Row defaults on INSERT if NOP)
    # weight uses TriState to support NULLIFY (set to None means use resource group's default_weight)
    weight: TriState[Decimal] = field(default_factory=TriState.nop)
    half_life_days: OptionalState[int] = field(default_factory=OptionalState.nop)
    lookback_days: OptionalState[int] = field(default_factory=OptionalState.nop)
    decay_unit_days: OptionalState[int] = field(default_factory=OptionalState.nop)
    resource_weights: OptionalState[ResourceSlot] = field(default_factory=OptionalState.nop)

    # Calculation snapshot (OptionalState - for UPDATE only)
    fair_share_factor: OptionalState[Decimal] = field(default_factory=OptionalState.nop)
    total_decayed_usage: OptionalState[ResourceSlot] = field(default_factory=OptionalState.nop)
    normalized_usage: OptionalState[Decimal] = field(default_factory=OptionalState.nop)
    lookback_start: OptionalState[date] = field(default_factory=OptionalState.nop)
    lookback_end: OptionalState[date] = field(default_factory=OptionalState.nop)

    @property
    @override
    def row_class(self) -> type[DomainFairShareRow]:
        return DomainFairShareRow

    @override
    def build_insert_values(self) -> dict[str, Any]:
        values: dict[str, Any] = {
            "resource_group": self.resource_group,
            "domain_name": self.domain_name,
        }
        # Spec fields
        self.weight.update_dict(values, "weight")
        self.half_life_days.update_dict(values, "half_life_days")
        self.lookback_days.update_dict(values, "lookback_days")
        self.decay_unit_days.update_dict(values, "decay_unit_days")
        self.resource_weights.update_dict(values, "resource_weights")
        # Calculation fields
        self.fair_share_factor.update_dict(values, "fair_share_factor")
        self.total_decayed_usage.update_dict(values, "total_decayed_usage")
        self.normalized_usage.update_dict(values, "normalized_usage")
        self.lookback_start.update_dict(values, "lookback_start")
        self.lookback_end.update_dict(values, "lookback_end")
        return values

    @override
    def build_update_values(self) -> dict[str, Any]:
        values: dict[str, Any] = {}
        # Spec fields (weight can be updated via mutation)
        self.weight.update_dict(values, "weight")
        # Calculation fields
        self.fair_share_factor.update_dict(values, "fair_share_factor")
        self.total_decayed_usage.update_dict(values, "total_decayed_usage")
        self.normalized_usage.update_dict(values, "normalized_usage")
        self.resource_weights.update_dict(values, "resource_weights")
        self.lookback_start.update_dict(values, "lookback_start")
        self.lookback_end.update_dict(values, "lookback_end")
        return values


@dataclass
class DomainFairShareBulkWeightUpserterSpec(UpserterSpec[DomainFairShareRow]):
    """Simplified upserter spec for bulk weight updates on DomainFairShareRow.

    Used with BulkUpserter for updating weights across multiple domains.
    Has fixed update column (weight only) to ensure consistent bulk operations.
    """

    resource_group: str
    domain_name: str
    weight: Decimal | None  # None means use resource group's default_weight

    @property
    @override
    def row_class(self) -> type[DomainFairShareRow]:
        return DomainFairShareRow

    @override
    def build_insert_values(self) -> dict[str, Any]:
        return {
            "resource_group": self.resource_group,
            "domain_name": self.domain_name,
            "weight": self.weight,
        }

    @override
    def build_update_values(self) -> dict[str, Any]:
        return {"weight": self.weight}


@dataclass
class ProjectFairShareUpserterSpec(UpserterSpec[ProjectFairShareRow]):
    """Upserter spec for ProjectFairShareRow.

    Unique constraint: (resource_group, project_id)

    - Identity fields are required (for unique constraint)
    - Spec fields use Row defaults on INSERT if NOP
    - Calculation fields are only updated if UPDATE
    """

    # Identity (required)
    resource_group: str
    project_id: uuid.UUID
    domain_name: str

    # Spec (uses Row defaults on INSERT if NOP)
    # weight uses TriState to support NULLIFY (set to None means use resource group's default_weight)
    weight: TriState[Decimal] = field(default_factory=TriState.nop)
    half_life_days: OptionalState[int] = field(default_factory=OptionalState.nop)
    lookback_days: OptionalState[int] = field(default_factory=OptionalState.nop)
    decay_unit_days: OptionalState[int] = field(default_factory=OptionalState.nop)
    resource_weights: OptionalState[ResourceSlot] = field(default_factory=OptionalState.nop)

    # Calculation snapshot (OptionalState - for UPDATE only)
    fair_share_factor: OptionalState[Decimal] = field(default_factory=OptionalState.nop)
    total_decayed_usage: OptionalState[ResourceSlot] = field(default_factory=OptionalState.nop)
    normalized_usage: OptionalState[Decimal] = field(default_factory=OptionalState.nop)
    lookback_start: OptionalState[date] = field(default_factory=OptionalState.nop)
    lookback_end: OptionalState[date] = field(default_factory=OptionalState.nop)

    @property
    @override
    def row_class(self) -> type[ProjectFairShareRow]:
        return ProjectFairShareRow

    @override
    def build_insert_values(self) -> dict[str, Any]:
        values: dict[str, Any] = {
            "resource_group": self.resource_group,
            "project_id": self.project_id,
            "domain_name": self.domain_name,
        }
        # Spec fields
        self.weight.update_dict(values, "weight")
        self.half_life_days.update_dict(values, "half_life_days")
        self.lookback_days.update_dict(values, "lookback_days")
        self.decay_unit_days.update_dict(values, "decay_unit_days")
        self.resource_weights.update_dict(values, "resource_weights")
        # Calculation fields
        self.fair_share_factor.update_dict(values, "fair_share_factor")
        self.total_decayed_usage.update_dict(values, "total_decayed_usage")
        self.normalized_usage.update_dict(values, "normalized_usage")
        self.lookback_start.update_dict(values, "lookback_start")
        self.lookback_end.update_dict(values, "lookback_end")
        return values

    @override
    def build_update_values(self) -> dict[str, Any]:
        values: dict[str, Any] = {}
        # Spec fields (weight can be updated via mutation)
        self.weight.update_dict(values, "weight")
        # Calculation fields
        self.fair_share_factor.update_dict(values, "fair_share_factor")
        self.total_decayed_usage.update_dict(values, "total_decayed_usage")
        self.normalized_usage.update_dict(values, "normalized_usage")
        self.resource_weights.update_dict(values, "resource_weights")
        self.lookback_start.update_dict(values, "lookback_start")
        self.lookback_end.update_dict(values, "lookback_end")
        return values


@dataclass
class ProjectFairShareBulkWeightUpserterSpec(UpserterSpec[ProjectFairShareRow]):
    """Simplified upserter spec for bulk weight updates on ProjectFairShareRow.

    Used with BulkUpserter for updating weights across multiple projects.
    Has fixed update column (weight only) to ensure consistent bulk operations.
    """

    resource_group: str
    project_id: uuid.UUID
    domain_name: str
    weight: Decimal | None  # None means use resource group's default_weight

    @property
    @override
    def row_class(self) -> type[ProjectFairShareRow]:
        return ProjectFairShareRow

    @override
    def build_insert_values(self) -> dict[str, Any]:
        return {
            "resource_group": self.resource_group,
            "project_id": self.project_id,
            "domain_name": self.domain_name,
            "weight": self.weight,
        }

    @override
    def build_update_values(self) -> dict[str, Any]:
        return {"weight": self.weight}


@dataclass
class UserFairShareUpserterSpec(UpserterSpec[UserFairShareRow]):
    """Upserter spec for UserFairShareRow.

    Unique constraint: (resource_group, user_uuid, project_id)

    - Identity fields are required (for unique constraint)
    - Spec fields use Row defaults on INSERT if NOP
    - Calculation fields are only updated if UPDATE
    """

    # Identity (required)
    resource_group: str
    user_uuid: uuid.UUID
    project_id: uuid.UUID
    domain_name: str

    # Spec (uses Row defaults on INSERT if NOP)
    # weight uses TriState to support NULLIFY (set to None means use resource group's default_weight)
    weight: TriState[Decimal] = field(default_factory=TriState.nop)
    half_life_days: OptionalState[int] = field(default_factory=OptionalState.nop)
    lookback_days: OptionalState[int] = field(default_factory=OptionalState.nop)
    decay_unit_days: OptionalState[int] = field(default_factory=OptionalState.nop)
    resource_weights: OptionalState[ResourceSlot] = field(default_factory=OptionalState.nop)

    # Calculation snapshot (OptionalState - for UPDATE only)
    fair_share_factor: OptionalState[Decimal] = field(default_factory=OptionalState.nop)
    total_decayed_usage: OptionalState[ResourceSlot] = field(default_factory=OptionalState.nop)
    normalized_usage: OptionalState[Decimal] = field(default_factory=OptionalState.nop)
    lookback_start: OptionalState[date] = field(default_factory=OptionalState.nop)
    lookback_end: OptionalState[date] = field(default_factory=OptionalState.nop)

    @property
    @override
    def row_class(self) -> type[UserFairShareRow]:
        return UserFairShareRow

    @override
    def build_insert_values(self) -> dict[str, Any]:
        values: dict[str, Any] = {
            "resource_group": self.resource_group,
            "user_uuid": self.user_uuid,
            "project_id": self.project_id,
            "domain_name": self.domain_name,
        }
        # Spec fields
        self.weight.update_dict(values, "weight")
        self.half_life_days.update_dict(values, "half_life_days")
        self.lookback_days.update_dict(values, "lookback_days")
        self.decay_unit_days.update_dict(values, "decay_unit_days")
        self.resource_weights.update_dict(values, "resource_weights")
        # Calculation fields
        self.fair_share_factor.update_dict(values, "fair_share_factor")
        self.total_decayed_usage.update_dict(values, "total_decayed_usage")
        self.normalized_usage.update_dict(values, "normalized_usage")
        self.lookback_start.update_dict(values, "lookback_start")
        self.lookback_end.update_dict(values, "lookback_end")
        return values

    @override
    def build_update_values(self) -> dict[str, Any]:
        values: dict[str, Any] = {}
        # Spec fields (weight can be updated via mutation)
        self.weight.update_dict(values, "weight")
        # Calculation fields
        self.fair_share_factor.update_dict(values, "fair_share_factor")
        self.total_decayed_usage.update_dict(values, "total_decayed_usage")
        self.normalized_usage.update_dict(values, "normalized_usage")
        self.resource_weights.update_dict(values, "resource_weights")
        self.lookback_start.update_dict(values, "lookback_start")
        self.lookback_end.update_dict(values, "lookback_end")
        return values


@dataclass
class UserFairShareBulkWeightUpserterSpec(UpserterSpec[UserFairShareRow]):
    """Simplified upserter spec for bulk weight updates on UserFairShareRow.

    Used with BulkUpserter for updating weights across multiple users.
    Has fixed update column (weight only) to ensure consistent bulk operations.
    """

    resource_group: str
    user_uuid: uuid.UUID
    project_id: uuid.UUID
    domain_name: str
    weight: Decimal | None  # None means use resource group's default_weight

    @property
    @override
    def row_class(self) -> type[UserFairShareRow]:
        return UserFairShareRow

    @override
    def build_insert_values(self) -> dict[str, Any]:
        return {
            "resource_group": self.resource_group,
            "user_uuid": self.user_uuid,
            "project_id": self.project_id,
            "domain_name": self.domain_name,
            "weight": self.weight,
        }

    @override
    def build_update_values(self) -> dict[str, Any]:
        return {"weight": self.weight}



from __future__ import annotations

import logging
from collections.abc import Iterable
from http import HTTPStatus

import aiohttp_cors
from aiohttp import web

from ai.backend.common.api_handlers import (
    APIResponse,
    BodyParam,
    PathParam,
    api_handler,
)
from ai.backend.common.dto.storage.request import GetVerificationResultReq
from ai.backend.common.dto.storage.response import GetVerificationResultResponse
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.artifact.types import ArtifactRevisionResponseData
from ai.backend.manager.dto.context import ProcessorsCtx
from ai.backend.manager.dto.request import (
    ApproveArtifactRevisionReq,
    CancelImportArtifactReq,
    CleanupArtifactsReq,
    GetArtifactRevisionReadmeReq,
    GetDownloadProgressReqPathParam,
    ImportArtifactsReq,
    RejectArtifactRevisionReq,
    UpdateArtifactReqBodyParam,
    UpdateArtifactReqPathParam,
)
from ai.backend.manager.dto.response import (
    ApproveArtifactRevisionResponse,
    ArtifactRevisionImportTask,
    CancelImportArtifactResponse,
    CleanupArtifactsResponse,
    GetArtifactRevisionReadmeResponse,
    GetDownloadProgressResponse,
    ImportArtifactsResponse,
    RejectArtifactRevisionResponse,
    UpdateArtifactResponse,
)
from ai.backend.manager.services.artifact.actions.update import UpdateArtifactAction
from ai.backend.manager.services.artifact_revision.actions.approve import (
    ApproveArtifactRevisionAction,
)
from ai.backend.manager.services.artifact_revision.actions.cancel_import import (
    CancelImportAction,
)
from ai.backend.manager.services.artifact_revision.actions.cleanup import (
    CleanupArtifactRevisionAction,
)
from ai.backend.manager.services.artifact_revision.actions.get_download_progress import (
    GetDownloadProgressAction,
)
from ai.backend.manager.services.artifact_revision.actions.get_readme import (
    GetArtifactRevisionReadmeAction,
)
from ai.backend.manager.services.artifact_revision.actions.get_verification_result import (
    GetArtifactRevisionVerificationResultAction,
)
from ai.backend.manager.services.artifact_revision.actions.import_revision import (
    ImportArtifactRevisionAction,
)
from ai.backend.manager.services.artifact_revision.actions.reject import (
    RejectArtifactRevisionAction,
)

from .auth import auth_required_for_method
from .types import CORSOptions, WebMiddleware

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class APIHandler:
    @auth_required_for_method
    @api_handler
    async def cleanup_artifacts(
        self,
        body: BodyParam[CleanupArtifactsReq],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """
        Clean up stored artifact revision data to free storage space.

        Removes the downloaded files for the specified artifact revisions and
        transitions them back to SCANNED status. The metadata remains, allowing
        the artifacts to be re-imported later if needed.

        Use this operation to manage storage usage by removing unused artifacts.
        """
        processors = processors_ctx.processors
        cleaned_revisions = []

        # Process each artifact revision sequentially
        # TODO: Optimize with asyncio.gather() for parallel processing
        for artifact_revision_id in body.parsed.artifact_revision_ids:
            action_result = await processors.artifact_revision.cleanup.wait_for_complete(
                CleanupArtifactRevisionAction(
                    artifact_revision_id=artifact_revision_id,
                )
            )
            cleaned_revisions.append(action_result.result)

        resp = CleanupArtifactsResponse(
            artifact_revisions=[
                ArtifactRevisionResponseData.from_revision_data(revision)
                for revision in cleaned_revisions
            ],
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def cancel_import_artifact(
        self,
        body: BodyParam[CancelImportArtifactReq],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """
        Cancel an in-progress artifact import operation.

        Stops the download process for the specified artifact revision and
        reverts its status back to SCANNED. The partially downloaded data is cleaned up.
        """
        processors = processors_ctx.processors
        action_result = await processors.artifact_revision.cancel_import.wait_for_complete(
            CancelImportAction(
                artifact_revision_id=body.parsed.artifact_revision_id,
            )
        )

        resp = CancelImportArtifactResponse(
            artifact_revision=ArtifactRevisionResponseData.from_revision_data(action_result.result),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def approve_artifact_revision(
        self,
        path: PathParam[ApproveArtifactRevisionReq],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """
        Approve an artifact revision for general use.

        Admin-only operation to approve artifact revisions, typically used
        in environments with approval workflows for artifact deployment.
        """
        processors = processors_ctx.processors
        action_result = await processors.artifact_revision.approve.wait_for_complete(
            ApproveArtifactRevisionAction(
                artifact_revision_id=path.parsed.artifact_revision_id,
            )
        )

        resp = ApproveArtifactRevisionResponse(
            artifact_revision=ArtifactRevisionResponseData.from_revision_data(action_result.result),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def reject_artifact_revision(
        self,
        path: PathParam[RejectArtifactRevisionReq],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """
        Reject an artifact revision, preventing its use.

        Admin-only operation to reject artifact revisions, typically used
        in environments with approval workflows for artifact deployment.
        """
        processors = processors_ctx.processors
        action_result = await processors.artifact_revision.reject.wait_for_complete(
            RejectArtifactRevisionAction(
                artifact_revision_id=path.parsed.artifact_revision_id,
            )
        )

        resp = RejectArtifactRevisionResponse(
            artifact_revision=ArtifactRevisionResponseData.from_revision_data(action_result.result),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def import_artifacts(
        self,
        body: BodyParam[ImportArtifactsReq],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """
        Import scanned artifact revisions from external registries.

        Downloads the actual files for the specified artifact revisions, transitioning
        them from SCANNED → PULLING → PULLED → AVAILABLE status.

        Returns background tasks that can be monitored for import progress.
        Once AVAILABLE, artifacts can be used by users in their sessions.
        """
        processors = processors_ctx.processors
        imported_revisions = []
        tasks = []

        # Process each artifact revision sequentially
        # TODO: Optimize with asyncio.gather() for parallel processing
        force = body.parsed.options.force if body.parsed.options else False
        for artifact_revision_id in body.parsed.artifact_revision_ids:
            # When using VFolderStorage (vfolder_id provided), store at root path
            storage_prefix = "/" if body.parsed.vfolder_id else None
            action_result = await processors.artifact_revision.import_revision.wait_for_complete(
                ImportArtifactRevisionAction(
                    artifact_revision_id=artifact_revision_id,
                    vfolder_id=body.parsed.vfolder_id,
                    storage_prefix=storage_prefix,
                    force=force,
                )
            )
            imported_revisions.append(action_result.result)
            tasks.append(
                ArtifactRevisionImportTask(
                    task_id=str(action_result.task_id),
                    artifact_revision=ArtifactRevisionResponseData.from_revision_data(
                        action_result.result
                    ),
                )
            )

        resp = ImportArtifactsResponse(tasks=tasks)
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def update_artifact(
        self,
        path: PathParam[UpdateArtifactReqPathParam],
        body: BodyParam[UpdateArtifactReqBodyParam],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """
        Update artifact metadata properties.

        Modifies artifact metadata such as readonly status and description.
        This operation does not affect the actual artifact files or revisions.
        """
        processors = processors_ctx.processors
        action_result = await processors.artifact.update.wait_for_complete(
            UpdateArtifactAction(
                updater=body.parsed.to_updater(path.parsed.artifact_id),
            )
        )

        resp = UpdateArtifactResponse(
            artifact=action_result.result,
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def get_artifact_revision_readme(
        self,
        path: PathParam[GetArtifactRevisionReadmeReq],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """
        Retrieve the README content for a specific artifact revision.

        Returns the README documentation associated with the artifact revision,
        which typically contains usage instructions and model information.
        """
        processors = processors_ctx.processors
        action_result = await processors.artifact_revision.get_readme.wait_for_complete(
            GetArtifactRevisionReadmeAction(
                artifact_revision_id=path.parsed.artifact_revision_id,
            )
        )

        resp = GetArtifactRevisionReadmeResponse(
            readme=action_result.readme_data.readme,
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def get_artifact_revision_verification_result(
        self,
        path: PathParam[GetVerificationResultReq],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """
        Retrieve the verification result for a specific artifact revision.

        Returns the verification result data associated with the artifact revision,
        which contains results from all verifiers that have been run on the artifact.
        """
        processors = processors_ctx.processors
        action_result = (
            await processors.artifact_revision.get_verification_result.wait_for_complete(
                GetArtifactRevisionVerificationResultAction(
                    artifact_revision_id=path.parsed.artifact_revision_id,
                )
            )
        )

        resp = GetVerificationResultResponse(
            verification_result=action_result.verification_result,
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    @auth_required_for_method
    @api_handler
    async def get_download_progress(
        self,
        path: PathParam[GetDownloadProgressReqPathParam],
        processors_ctx: ProcessorsCtx,
    ) -> APIResponse:
        """
        Retrieve download progress for an artifact revision.

        Returns detailed download progress information including artifact-level
        and file-level progress data for the specified artifact revision.
        Supports both local and remote download progress when delegation is enabled.
        """
        processors = processors_ctx.processors
        action_result = await processors.artifact_revision.get_download_progress.wait_for_complete(
            GetDownloadProgressAction(
                artifact_revision_id=path.parsed.artifact_revision_id,
            )
        )

        resp = GetDownloadProgressResponse(
            download_progress=action_result.download_progress,
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["api_versions"] = (1, 2, 3, 4, 5)
    app["prefix"] = "artifacts"
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    api_handler = APIHandler()

    cors.add(app.router.add_route("POST", "/revisions/cleanup", api_handler.cleanup_artifacts))
    cors.add(
        app.router.add_route(
            "POST",
            "/revisions/{artifact_revision_id}/approval",
            api_handler.approve_artifact_revision,
        )
    )
    cors.add(
        app.router.add_route(
            "POST",
            "/revisions/{artifact_revision_id}/rejection",
            api_handler.reject_artifact_revision,
        )
    )
    cors.add(app.router.add_route("POST", "/task/cancel", api_handler.cancel_import_artifact))
    cors.add(app.router.add_route("POST", "/import", api_handler.import_artifacts))
    cors.add(app.router.add_route("PATCH", "/{artifact_id}", api_handler.update_artifact))
    cors.add(
        app.router.add_route(
            "GET",
            "/revisions/{artifact_revision_id}/readme",
            api_handler.get_artifact_revision_readme,
        )
    )
    cors.add(
        app.router.add_route(
            "GET",
            "/revisions/{artifact_revision_id}/verification-result",
            api_handler.get_artifact_revision_verification_result,
        )
    )
    cors.add(
        app.router.add_route(
            "GET",
            "/revisions/{artifact_revision_id}/download-progress",
            api_handler.get_download_progress,
        )
    )

    return app, []



"""UpdaterSpec implementations for vfolder repository."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, override

from ai.backend.manager.models.vfolder import VFolderPermission, VFolderRow
from ai.backend.manager.repositories.base.updater import UpdaterSpec
from ai.backend.manager.types import OptionalState


@dataclass
class VFolderAttributeUpdaterSpec(UpdaterSpec[VFolderRow]):
    """UpdaterSpec for vfolder attribute updates."""

    name: OptionalState[str] = field(default_factory=OptionalState.nop)
    cloneable: OptionalState[bool] = field(default_factory=OptionalState.nop)
    mount_permission: OptionalState[VFolderPermission] = field(default_factory=OptionalState.nop)

    @property
    @override
    def row_class(self) -> type[VFolderRow]:
        return VFolderRow

    @override
    def build_values(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.name.update_dict(to_update, "name")
        self.cloneable.update_dict(to_update, "cloneable")
        self.mount_permission.update_dict(to_update, "permission")
        return to_update



from __future__ import annotations

import logging
from collections.abc import Mapping
from typing import cast
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy.exc import DBAPIError
from sqlalchemy.ext.asyncio import AsyncSession as SASession
from sqlalchemy.orm import selectinload

from ai.backend.common.bgtask.reporter import ProgressReporter
from ai.backend.common.docker import ImageRef
from ai.backend.common.exception import UnknownImageReference
from ai.backend.common.types import ImageAlias, ImageID
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.data.image.types import (
    ImageAliasData,
    ImageAliasListResult,
    ImageData,
    ImageDataWithDetails,
    ImageListResult,
    ImageStatus,
    RescanImagesResult,
    ResourceLimitInput,
)
from ai.backend.manager.errors.image import (
    AliasImageActionDBError,
    AliasImageActionValueError,
    ImageAliasNotFound,
    ImageNotFound,
    ModifyImageActionValueError,
    PurgeImageActionByIdObjectDBError,
    RegistryNotFoundForImage,
)
from ai.backend.manager.models.container_registry import ContainerRegistryRow
from ai.backend.manager.models.image import (
    ImageAliasRow,
    ImageIdentifier,
    ImageRow,
    rescan_images,
    scan_single_image,
)
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.repositories.base import BatchQuerier, Creator, execute_batch_querier
from ai.backend.manager.repositories.base.creator import execute_creator
from ai.backend.manager.repositories.base.updater import Updater, execute_updater
from ai.backend.manager.repositories.image.creators import ImageAliasCreatorSpec

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ImageDBSource:
    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    async def fetch_image_by_identifiers(
        self, identifiers: list[ImageAlias | ImageRef | ImageIdentifier]
    ) -> ImageData:
        """
        Fetches an image from database by its identifiers, which can be a combination of
        ImageAlias, ImageRef, or ImageIdentifier.
        Returns an ImageData object.
        Raises Exception if the image cannot be found.
        """
        async with self._db.begin_readonly_session_read_committed() as session:
            row = await self._resolve_image(session, identifiers)
            return row.to_dataclass()

    async def fetch_images_batch(
        self, identifier_lists: list[list[ImageIdentifier]]
    ) -> list[ImageData]:
        """
        Fetches multiple images from database by their identifiers in a single database session.
        Returns a list of ImageData objects.
        More efficient than multiple individual fetch operations.
        """
        async with self._db.begin_readonly_session() as session:
            rows: list[ImageRow] = []
            for identifiers in identifier_lists:
                row = await self._resolve_image(
                    session, cast(list[ImageAlias | ImageRef | ImageIdentifier], identifiers)
                )
                rows.append(row)
            return [row.to_dataclass() for row in rows]

    async def _resolve_image(
        self,
        session: SASession,
        identifiers: list[ImageAlias | ImageRef | ImageIdentifier],
    ) -> ImageRow:
        return await ImageRow.resolve(session, identifiers)

    async def _get_image_by_id(
        self,
        session: SASession,
        image_id: UUID,
        load_aliases: bool = False,
        status_filter: list[ImageStatus] | None = None,
    ) -> ImageRow:
        """
        Private method to get an image by ID using an existing session.
        Returns None if image is not found.
        """
        row = await ImageRow.get(
            session, image_id, load_aliases=load_aliases, filter_by_statuses=status_filter
        )
        if row is None:
            raise ImageNotFound()
        return row

    async def _get_image_alias_by_name(self, session: SASession, alias: str) -> ImageAliasRow:
        """
        Private method to get an image alias by name using an existing session.
        """
        image_alias_row = await session.scalar(
            sa.select(ImageAliasRow).where(ImageAliasRow.alias == alias),
        )
        if not image_alias_row:
            raise ImageAliasNotFound(f"Image alias '{alias}' not found.")
        return image_alias_row

    async def query_images_by_canonicals(
        self,
        canonicals: list[str],
        status_filter: list[ImageStatus] | None = None,
    ) -> dict[ImageID, ImageDataWithDetails]:
        """
        Deprecated. Use query_images_by_ids instead.
        """
        query = (
            sa.select(ImageRow)
            .where(ImageRow.name.in_(canonicals))
            .options(selectinload(ImageRow.aliases))
        )
        if status_filter:
            query = query.where(ImageRow.status.in_(status_filter))

        async with self._db.begin_readonly_session_read_committed() as session:
            result = await session.execute(query)
            image_rows = list(result.scalars().all())
            return {ImageID(row.id): row.to_detailed_dataclass() for row in image_rows}

    async def query_image_details_by_identifier(
        self,
        identifier: ImageIdentifier,
        status_filter: list[ImageStatus] | None = None,
    ) -> ImageDataWithDetails:
        """
        Deprecated. Use query_image_details_by_id instead.
        """
        try:
            async with self._db.begin_readonly_session_read_committed() as session:
                image_row = await ImageRow.resolve(
                    session,
                    [
                        identifier,
                        ImageAlias(identifier.canonical),
                    ],
                    filter_by_statuses=status_filter,
                )
        except UnknownImageReference as e:
            raise ImageNotFound from e
        return image_row.to_detailed_dataclass()

    async def query_image_details_by_id(
        self,
        image_id: UUID,
        load_aliases: bool = False,
        status_filter: list[ImageStatus] | None = None,
    ) -> ImageDataWithDetails:
        async with self._db.begin_readonly_session_read_committed() as session:
            try:
                row: ImageRow = await self._get_image_by_id(
                    session, image_id, load_aliases, status_filter
                )
            except UnknownImageReference as e:
                raise ImageNotFound() from e
            return row.to_detailed_dataclass()

    async def query_all_images(
        self, status_filter: list[ImageStatus] | None = None
    ) -> Mapping[ImageID, ImageDataWithDetails]:
        async with self._db.begin_readonly_session_read_committed() as session:
            rows = await ImageRow.list(session, load_aliases=True, filter_by_statuses=status_filter)
            return {ImageID(row.id): row.to_detailed_dataclass() for row in rows}

    async def mark_image_deleted(
        self,
        identifiers: list[ImageAlias | ImageRef | ImageIdentifier],
    ) -> ImageData:
        """
        Deprecated. Use mark_image_deleted_by_id instead.
        """
        async with self._db.begin_session() as session:
            row = await self._resolve_image(session, identifiers)
            await row.mark_as_deleted(session)
            return row.to_dataclass()

    async def mark_image_deleted_by_id(
        self,
        image_id: UUID,
    ) -> ImageData:
        """
        Marks an image record as deleted by its ID in the database.
        """
        async with self._db.begin_session() as session:
            image_row = await self._get_image_by_id(session, image_id)
            await image_row.mark_as_deleted(session)
            return image_row.to_dataclass()

    async def fetch_image_by_id(self, image_id: UUID, load_aliases: bool = False) -> ImageData:
        """
        Fetches an image from database by ID.
        Raises ImageNotFound if image doesn't exist.
        """
        async with self._db.begin_readonly_session_read_committed() as session:
            image_row = await self._get_image_by_id(session, image_id, load_aliases)
            return image_row.to_dataclass()

    async def validate_image_ownership(self, image_id: UUID, user_id: UUID) -> bool:
        """
        Checks if user owns the image.
        Returns True if user owns the image, False otherwise.
        Raises ImageNotFound if image doesn't exist.
        """
        async with self._db.begin_readonly_session_read_committed() as session:
            image_row = await self._get_image_by_id(session, image_id)
            return image_row.is_owned_by(user_id)

    async def insert_image_alias(
        self, alias: str, image_canonical: str, architecture: str
    ) -> tuple[UUID, ImageAliasData]:
        """
        Deprecated. Use insert_image_alias_by_id instead.
        """
        try:
            async with self._db.begin_session() as session:
                image_row = await ImageRow.resolve(
                    session, [ImageIdentifier(image_canonical, architecture)]
                )
                image_alias = ImageAliasRow(alias=alias, image_id=image_row.id)
                image_row.aliases.append(image_alias)
                row_id = image_row.id
                alias_data = ImageAliasData(id=image_alias.id, alias=image_alias.alias or "")
            return row_id, alias_data
        except ValueError as e:
            raise AliasImageActionValueError from e
        except DBAPIError as e:
            raise AliasImageActionDBError(str(e)) from e

    async def query_image_alias(self, alias: str) -> ImageAliasData:
        async with self._db.begin_readonly_session_read_committed() as session:
            row = await self._get_image_alias_by_name(session, alias)
            return ImageAliasData(id=row.id, alias=row.alias or "")

    async def remove_image_alias(self, alias: str) -> tuple[UUID, ImageAliasData]:
        async with self._db.begin_session() as session:
            existing_alias = await self._get_image_alias_by_name(session, alias)
            image_id = existing_alias.image_id
            alias_data = ImageAliasData(id=existing_alias.id, alias=existing_alias.alias or "")
            await session.delete(existing_alias)
        return image_id, alias_data

    async def scan_and_upsert_image(
        self, image_canonical: str, architecture: str
    ) -> RescanImagesResult:
        """
        Deprecated. Use scan_images_by_ids instead.
        """

        async with self._db.begin_session() as session:
            # Resolve the image first
            image_row = await self._resolve_image(
                session, [ImageIdentifier(image_canonical, architecture)]
            )

            # Get the registry info
            registry_parts = []
            if image_row.registry:
                registry_parts.append(image_row.registry)
            if image_row.project:
                registry_parts.append(image_row.project)
            registry_key = "/".join(registry_parts) if registry_parts else ""

            # Get the registry row
            registry_row = await session.get(ContainerRegistryRow, image_row.registry_id)
            if not registry_row:
                raise RegistryNotFoundForImage(f"Registry not found for image {image_canonical}")

            # Call the original scan function
            return await scan_single_image(self._db, registry_key, registry_row, image_canonical)

    async def remove_tag_from_registry(self, image_id: UUID) -> ImageData:
        async with self._db.begin_readonly_session() as session:
            image_row = await self._get_image_by_id(session, image_id, load_aliases=True)
            if not image_row:
                raise ImageNotFound()
            await image_row.untag_image_from_registry(self._db, session)
            return image_row.to_dataclass()

    async def modify_image_properties(self, updater: Updater[ImageRow]) -> ImageData:
        try:
            async with self._db.begin_session() as session:
                result = await execute_updater(session, updater)
                if result is None:
                    raise ImageNotFound(f"Image not found (id:{updater.pk_value})")
                return result.row.to_dataclass()
        except (ValueError, DBAPIError) as e:
            raise ModifyImageActionValueError from e

    async def clear_image_resource_limits(
        self, image_canonical: str, architecture: str
    ) -> ImageData:
        """
        Deprecated. Use clear_image_resource_limits_by_id instead.
        """
        async with self._db.begin_session() as session:
            image_row = await ImageRow.resolve(
                session, [ImageIdentifier(image_canonical, architecture)]
            )
            image_row._resources = {}
            return image_row.to_dataclass()

    async def insert_image_alias_by_id(self, creator: Creator[ImageAliasRow]) -> ImageAliasData:
        """
        Creates an image alias using the Creator pattern.
        """
        spec = cast(ImageAliasCreatorSpec, creator.spec)
        try:
            async with self._db.begin_session() as session:
                # Validate that the image exists
                await self._get_image_by_id(session, spec.image_id)
                result = await execute_creator(session, creator)
                return ImageAliasData(id=result.row.id, alias=result.row.alias or "")
        except ValueError as e:
            raise AliasImageActionValueError from e
        except DBAPIError as e:
            raise AliasImageActionDBError(str(e)) from e

    async def query_images_by_ids(
        self,
        image_ids: list[UUID],
        status_filter: list[ImageStatus] | None = None,
    ) -> dict[ImageID, ImageDataWithDetails]:
        """
        Queries images by their IDs with optional status filter.
        Returns a dictionary mapping ImageID to ImageDataWithDetails.
        """
        if not image_ids:
            return {}

        query = (
            sa.select(ImageRow)
            .where(ImageRow.id.in_(image_ids))
            .options(selectinload(ImageRow.aliases))
        )
        if status_filter:
            query = query.where(ImageRow.status.in_(status_filter))

        async with self._db.begin_readonly_session_read_committed() as session:
            result = await session.execute(query)
            image_rows = list(result.scalars().all())
            return {ImageID(row.id): row.to_detailed_dataclass() for row in image_rows}

    async def clear_image_resource_limits_by_id(self, image_id: UUID) -> ImageData:
        """
        Clears image resource limits by image ID.
        """
        async with self._db.begin_session() as session:
            image_row = await self._get_image_by_id(session, image_id)
            image_row._resources = {}
            return image_row.to_dataclass()

    async def set_image_resource_limit_by_id(
        self,
        image_id: UUID,
        resource_limit: ResourceLimitInput,
    ) -> ImageData:
        """
        Sets resource limit for an image by its ID.
        """
        async with self._db.begin_session() as session:
            image_row = await self._get_image_by_id(session, image_id)
            resources = dict(image_row._resources) if image_row._resources else {}

            if resource_limit.slot_name not in resources:
                resources[resource_limit.slot_name] = {"min": None, "max": None}

            if resource_limit.min_value is not None:
                resources[resource_limit.slot_name]["min"] = str(resource_limit.min_value)
            if resource_limit.max_value is not None:
                resources[resource_limit.slot_name]["max"] = str(resource_limit.max_value)

            image_row._resources = resources
            return image_row.to_dataclass()

    async def remove_image_and_aliases(
        self,
        image_id: UUID,
    ) -> ImageData:
        """
        Removes an image record and all its aliases from the database.
        """
        try:
            async with self._db.begin_session() as session:
                image_row = await self._get_image_by_id(session, image_id, load_aliases=True)
                data = image_row.to_dataclass()
                for alias in image_row.aliases:
                    await session.delete(alias)
                await session.delete(image_row)
            return data
        except DBAPIError as e:
            raise PurgeImageActionByIdObjectDBError(str(e)) from e

    async def search_images(self, querier: BatchQuerier) -> ImageListResult:
        """
        Search images using a batch querier with conditions, pagination, and ordering.
        Returns ImageListResult with items and pagination info.
        """
        async with self._db.begin_readonly_session() as db_sess:
            query = sa.select(ImageRow).options(selectinload(ImageRow.aliases))
            result = await execute_batch_querier(db_sess, query, querier)
            items = [row.ImageRow.to_dataclass() for row in result.rows]
            return ImageListResult(
                items=items,
                total_count=result.total_count,
                has_next_page=result.has_next_page,
                has_previous_page=result.has_previous_page,
            )

    async def search_aliases(self, querier: BatchQuerier) -> ImageAliasListResult:
        """
        Search image aliases using a batch querier with conditions, pagination, and ordering.
        Returns ImageAliasListResult with items and pagination info.
        """
        async with self._db.begin_readonly_session() as db_sess:
            query = sa.select(ImageAliasRow)
            result = await execute_batch_querier(db_sess, query, querier)
            items = [row.ImageAliasRow.to_dataclass() for row in result.rows]
            image_ids = [ImageID(row.ImageAliasRow.image_id) for row in result.rows]
            return ImageAliasListResult(
                items=items,
                image_ids=image_ids,
                total_count=result.total_count,
                has_next_page=result.has_next_page,
                has_previous_page=result.has_previous_page,
            )

    async def rescan_images(
        self,
        registry_or_image: str | None = None,
        project: str | None = None,
        *,
        reporter: ProgressReporter | None = None,
    ) -> RescanImagesResult:
        """
        Rescan container registries and update images table.
        Wraps the rescan_images function from models/image/row.py.
        """
        return await rescan_images(
            self._db,
            registry_or_image,
            project,
            reporter=reporter,
        )



"""
Agent-related exceptions.
"""

from __future__ import annotations

from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)
from ai.backend.common.types import AgentId


class AgentConnectionUnavailable(BackendAIError, web.HTTPServiceUnavailable):
    """Raised when an Agent connection is unavailable."""

    error_type = "https://api.backend.ai/probs/agent-connection-unavailable"
    error_title = "Agent connection unavailable."

    def __init__(self, agent_id: AgentId, failure_reason: str) -> None:
        self.agent_id = agent_id
        self.failure_reason = failure_reason
        super().__init__(f"Agent {agent_id} connection unavailable: {failure_reason}")

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.AGENT,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.UNAVAILABLE,
        )



"""Allocation-related data types for scheduling results."""

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING, Any
from uuid import UUID

from ai.backend.common.types import (
    AccessKey,
    AgentId,
    ClusterMode,
    ResourceSlot,
    SessionId,
    SessionTypes,
)

from .workload import SessionWorkload

if TYPE_CHECKING:
    from ai.backend.manager.sokovan.scheduler.provisioner.selectors.selector import (
        AgentSelection,
    )


@dataclass
class SchedulingPredicate:
    """Represents a scheduling predicate (passed or failed)."""

    # Name of the component that generated this predicate
    name: str
    # Message describing the result
    msg: str

    def serialize(self) -> dict[str, str]:
        """Convert to dictionary for JSON serialization."""
        return {"name": self.name, "msg": self.msg}


@dataclass
class KernelAllocation:
    """Represents an allocation decision for a single kernel."""

    # Unique identifier of the kernel
    kernel_id: UUID
    # Identifier of the agent where this kernel will be allocated
    agent_id: AgentId
    # Network address of the agent
    agent_addr: str
    # Scaling group that the agent belongs to
    scaling_group: str
    # Host ports allocated for this kernel (empty set if none)
    allocated_host_ports: set[int] = field(default_factory=set)


@dataclass
class AgentAllocation:
    """Represents resource allocation to a specific agent."""

    # Identifier of the agent
    agent_id: AgentId
    # List of resource slots allocated to this agent
    allocated_slots: list[ResourceSlot]


@dataclass
class SessionAllocation:
    """Represents an allocation decision for a session with all its kernels."""

    # Unique identifier of the session
    session_id: SessionId
    # Type of the session (INTERACTIVE, BATCH, INFERENCE)
    session_type: SessionTypes
    # Cluster mode of the session (SINGLE_NODE or MULTI_NODE)
    cluster_mode: ClusterMode
    # Scaling group that the session belongs to
    scaling_group: str
    # List of kernel allocations for this session
    kernel_allocations: list[KernelAllocation]
    # List of agent allocations for this session
    agent_allocations: list[AgentAllocation]
    # Keypair associated with the session
    access_key: AccessKey
    # Phases that passed during scheduling
    passed_phases: list[SchedulingPredicate] = field(default_factory=list)
    # Phases that failed during scheduling (normally empty for successful allocations)
    failed_phases: list[SchedulingPredicate] = field(default_factory=list)

    @classmethod
    def from_agent_selections(
        cls,
        session_workload: SessionWorkload,
        selections: list[AgentSelection],
        scaling_group: str,
    ) -> SessionAllocation:
        """
        Build a SessionAllocation from agent selection results.

        :param session_workload: The original session workload
        :param selections: List of agent selection results
        :param scaling_group: The scaling group name
        :param access_key: The access key associated with the session
        :return: SessionAllocation with kernel and agent allocations
        """
        kernel_allocations: list[KernelAllocation] = []
        agent_allocation_map: dict[AgentId, AgentAllocation] = {}

        for selection in selections:
            resource_req = selection.resource_requirements
            selected_agent = selection.selected_agent

            # Track resource allocation for this agent
            if selected_agent.agent_id not in agent_allocation_map:
                agent_allocation_map[selected_agent.agent_id] = AgentAllocation(
                    agent_id=selected_agent.agent_id,
                    allocated_slots=[],
                )
            agent_allocation_map[selected_agent.agent_id].allocated_slots.append(
                resource_req.requested_slots
            )

            # Create kernel allocations
            for kernel_id in resource_req.kernel_ids:
                kernel_allocations.append(
                    KernelAllocation(
                        kernel_id=kernel_id,
                        agent_id=selected_agent.agent_id,
                        agent_addr=selected_agent.agent_addr,
                        scaling_group=selected_agent.scaling_group,
                    )
                )

        # Create session allocation
        agent_allocations = list(agent_allocation_map.values())

        return cls(
            session_id=session_workload.session_id,
            session_type=session_workload.session_type,
            cluster_mode=session_workload.cluster_mode,
            scaling_group=scaling_group,
            kernel_allocations=kernel_allocations,
            agent_allocations=agent_allocations,
            access_key=session_workload.access_key,
        )

    def unique_agent_ids(self) -> list[AgentId]:
        """Extract unique agent IDs from kernel allocations."""
        return list({
            kernel_alloc.agent_id
            for kernel_alloc in self.kernel_allocations
            if kernel_alloc.agent_id is not None
        })


@dataclass
class SchedulingFailure:
    """Information about a scheduling failure for status updates.

    Maintains compatibility with frontend scheduler JSON structure:
    {
        failed_predicates: Array<{name: string, msg?: string}>,
        passed_predicates: Array<{name: string}>,
        retries: number,
        last_try: string,
        msg?: string
    }
    """

    session_id: SessionId
    passed_phases: list[SchedulingPredicate] = field(default_factory=list)
    failed_phases: list[SchedulingPredicate] = field(default_factory=list)
    last_try: datetime | None = None
    msg: str | None = None

    def to_status_data(self, current_retries: int) -> dict[str, Any]:
        """Convert failure to status data dictionary for storage."""
        return {
            "passed_predicates": [p.serialize() for p in self.passed_phases],
            "failed_predicates": [p.serialize() for p in self.failed_phases],
            "retries": current_retries + 1,
            "last_try": self.last_try.isoformat() if self.last_try else None,
            "msg": self.msg,
        }


@dataclass
class AllocationBatch:
    """Bundle of session allocations and scheduling failures for batch processing."""

    # Successful allocations to process
    allocations: list[SessionAllocation]
    # Failed scheduling attempts to update status for
    failures: list[SchedulingFailure]

    def get_agent_ids(self) -> set[AgentId]:
        """Extract all agent IDs from allocations for efficient pre-fetching."""
        agent_ids: set[AgentId] = set()
        for allocation in self.allocations:
            for agent_alloc in allocation.agent_allocations:
                agent_ids.add(agent_alloc.agent_id)
        return agent_ids



"""
Configuration management CLI commands for Backend.AI Manager.

This module provides CLI commands for generating and managing configuration files.
"""

import logging
import pathlib

import click

from ai.backend.common.cli import LazyGroup
from ai.backend.common.configs.generator import (
    GeneratorConfig,
    TOMLGenerator,
)
from ai.backend.common.meta import ConfigEnvironment
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.config.unified import ManagerUnifiedConfig

from .context import CLIContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


@click.group()
def cli() -> None:
    """Configuration management commands."""
    pass


@cli.command()
@click.option(
    "-o",
    "--output",
    type=click.Path(
        file_okay=True,
        dir_okay=False,
        writable=True,
        path_type=pathlib.Path,
    ),
    default="./configs/manager/sample.toml",
    help="Output path for the generated sample configuration file. (default: sample.toml)",
)
@click.option(
    "-e",
    "--env",
    type=click.Choice(["local", "prod"]),
    default="prod",
    help="Target environment for example values. (default: prod)",
)
@click.option(
    "--overwrite",
    is_flag=True,
    help="Overwrite the output file if it already exists.",
)
@click.option(
    "--unmask-secrets",
    is_flag=True,
    help="Show actual secret values instead of masking them.",
)
@click.pass_obj
def generate_sample(
    _cli_ctx: CLIContext,
    output: pathlib.Path,
    env: str,
    overwrite: bool,
    unmask_secrets: bool,
) -> None:
    """
    Generate a sample configuration file from the ManagerUnifiedConfig schema.

    This command creates a TOML configuration file with environment-specific examples
    (local or prod), proper secret masking, and comprehensive documentation from
    BackendAIConfigMeta annotations.

    Features:
    - Environment-specific examples (--env local|prod)
    - Secret field masking (use --unmask-secrets to show actual values)
    - Comprehensive field documentation from BackendAIConfigMeta
    """
    if output.exists() and not overwrite:
        click.echo(
            f"Error: Output file '{output}' already exists. Use --overwrite to replace it.",
            err=True,
        )
        return

    header_comment = f"""
Backend.AI Manager Configuration ({env.upper()} Environment)

This is a sample configuration file for the Backend.AI Manager.
All configuration options are documented with their descriptions,
default values, and environment-specific examples.

Generated using BackendAIConfigMeta annotations.
"""

    config_env = ConfigEnvironment.LOCAL if env == "local" else ConfigEnvironment.PROD

    try:
        generator_config = GeneratorConfig(
            mask_secrets=not unmask_secrets,
            include_version_comments=True,
        )
        generator = TOMLGenerator(env=config_env, config=generator_config)
        generator.generate_to_file(ManagerUnifiedConfig, output, header=header_comment.strip())
        log.info(f"Sample configuration file generated successfully: {output}")
    except Exception as e:
        raise click.ClickException(f"Failed to generate sample configuration: {e}") from e


@cli.group(cls=LazyGroup, import_name="ai.backend.manager.cli.config_migrate:cli")
def migrate() -> None:
    """Migrate legacy configuration fields."""


if __name__ == "__main__":
    cli()



from __future__ import annotations

import logging
from collections.abc import Iterable, Mapping, Sequence
from decimal import Decimal
from typing import (
    TYPE_CHECKING,
    Any,
    Self,
    cast,
    overload,
)
from uuid import UUID

import graphene
import graphene_federation
import graphql
import sqlalchemy as sa
from dateutil.parser import parse as dtparse
from graphql import Undefined, UndefinedType
from sqlalchemy.orm import selectinload

from ai.backend.common.bgtask.reporter import ProgressReporter
from ai.backend.common.docker import ImageRef, KernelFeatures, LabelName
from ai.backend.common.types import (
    AgentId,
    DispatchResult,
    ImageID,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.bgtask.tasks.purge_images import (
    PurgeAgentSpec,
    PurgeImagesManifest,
    PurgeImageSpec,
)
from ai.backend.manager.bgtask.types import ManagerBgtaskName
from ai.backend.manager.data.image.types import (
    ImageData,
    ImageStatus,
    ImageType,
    ImageWithAgentInstallStatus,
)
from ai.backend.manager.defs import DEFAULT_IMAGE_ARCH
from ai.backend.manager.models.image import (
    ImageIdentifier,
    ImageLoadFilter,
    ImageRow,
    get_permission_ctx,
)
from ai.backend.manager.models.minilang import EnumFieldItem
from ai.backend.manager.models.minilang.ordering import ColumnMapType, QueryOrderParser
from ai.backend.manager.models.minilang.queryfilter import (
    FieldSpecType,
    QueryFilterParser,
)
from ai.backend.manager.models.rbac import ScopeType
from ai.backend.manager.models.rbac.context import ClientContext
from ai.backend.manager.models.rbac.permission_defs import ImagePermission
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.repositories.image.updaters import ImageUpdaterSpec
from ai.backend.manager.services.container_registry.actions.clear_images import ClearImagesAction
from ai.backend.manager.services.container_registry.actions.load_all_container_registries import (
    LoadAllContainerRegistriesAction,
)
from ai.backend.manager.services.container_registry.actions.load_container_registries import (
    LoadContainerRegistriesAction,
)
from ai.backend.manager.services.container_registry.actions.rescan_images import RescanImagesAction
from ai.backend.manager.services.image.actions.alias_image import AliasImageAction
from ai.backend.manager.services.image.actions.clear_image_custom_resource_limit import (
    ClearImageCustomResourceLimitAction,
)
from ai.backend.manager.services.image.actions.dealias_image import DealiasImageAction
from ai.backend.manager.services.image.actions.forget_image import (
    ForgetImageAction,
    ForgetImageByIdAction,
)
from ai.backend.manager.services.image.actions.get_all_images import GetAllImagesAction
from ai.backend.manager.services.image.actions.get_image_installed_agents import (
    GetImageInstalledAgentsAction,
)
from ai.backend.manager.services.image.actions.get_images import (
    GetImageByIdAction,
    GetImageByIdentifierAction,
    GetImagesByCanonicalsAction,
)
from ai.backend.manager.services.image.actions.modify_image import (
    ModifyImageAction,
)
from ai.backend.manager.services.image.actions.purge_images import PurgeImageByIdAction
from ai.backend.manager.services.image.actions.untag_image_from_registry import (
    UntagImageFromRegistryAction,
)
from ai.backend.manager.types import OptionalState, TriState

from .base import (
    BigInt,
    FilterExprArg,
    ImageRefType,
    KVPair,
    KVPairInput,
    OrderExprArg,
    ResourceLimit,
    ResourceLimitInput,
    batch_multiresult_in_scalar_stream,
    extract_object_uuid,
    generate_sql_info_for_gql_connection,
)
from .gql_relay import AsyncNode, Connection, ConnectionResolverResult, ResolvedGlobalID

if TYPE_CHECKING:
    from .schema import GraphQueryContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

__all__ = (
    "AliasImage",
    "ClearImages",
    "DealiasImage",
    "ForgetImage",
    "ForgetImageById",
    "Image",
    "ImageNode",
    "ModifyImage",
    "PreloadImage",
    "PurgeImageById",
    "RescanImages",
    "UntagImageFromRegistry",
)

_queryfilter_fieldspec: FieldSpecType = {
    "id": ("id", None),
    "name": ("name", None),
    "namespace": ("image", None),
    "tag": ("tag", None),
    "status": (EnumFieldItem("status", ImageStatus), None),
    "project": ("project", None),
    "image": ("image", None),
    "base_image_name": ("image", None),
    "created_at": ("created_at", dtparse),
    "registry": ("registry", None),
    "registry_id": ("registry_id", None),
    "architecture": ("architecture", None),
    "is_local": ("is_local", None),
    "type": (EnumFieldItem("type", ImageType), None),
    "accelerators": ("accelerators", None),
}

_queryorder_colmap: ColumnMapType = {
    "id": ("id", None),
    "name": ("name", None),
    "namespace": ("image", None),
    "tag": ("tag", None),
    "status": ("status", None),
    "project": ("project", None),
    "image": ("image", None),
    "base_image_name": ("image", None),
    "created_at": ("created_at", None),
    "registry": ("registry", None),
    "registry_id": ("registry_id", None),
    "architecture": ("architecture", None),
    "is_local": ("is_local", None),
    "type": ("session_type", None),
    "accelerators": ("accelerators", None),
}

ImageStatusType = graphene.Enum.from_enum(ImageStatus, description="Added in 25.4.0.")
ImageTypeEnum = graphene.Enum.from_enum(ImageType, description="Added in 25.12.0.")


class Image(graphene.ObjectType):  # type: ignore[misc]
    id = graphene.UUID()
    name = graphene.String(deprecation_reason="Deprecated since 24.12.0. use `namespace` instead")
    namespace = graphene.String(description="Added in 24.12.0.")
    base_image_name = graphene.String(description="Added in 24.12.0.")
    project = graphene.String(description="Added in 24.03.10.")
    humanized_name = graphene.String()
    tag = graphene.String()
    tags = graphene.List(KVPair, description="Added in 24.12.0.")
    version = graphene.String(description="Added in 24.12.0.")
    registry = graphene.String()
    architecture = graphene.String()
    is_local = graphene.Boolean()
    digest = graphene.String()
    labels = graphene.List(KVPair)
    aliases = graphene.List(graphene.String)
    size_bytes = BigInt()
    status = graphene.String(description="Added in 25.4.0.")
    resource_limits = graphene.List(ResourceLimit)
    supported_accelerators = graphene.List(graphene.String)
    installed = graphene.Boolean()
    installed_agents = graphene.List(graphene.String)
    # legacy field
    hash = graphene.String()

    @classmethod
    def from_image_with_agent_install_status(cls, data: ImageWithAgentInstallStatus) -> Self:
        return cls(
            id=data.image.id,
            name=data.image.name,
            namespace=data.image.namespace,
            base_image_name=data.image.base_image_name,
            project=data.image.project,
            humanized_name=data.image.humanized_name,
            tag=data.image.tag,
            tags=[KVPair(key=kvpair.key, value=kvpair.value) for kvpair in data.image.tags],
            version=data.image.version,
            registry=data.image.registry,
            architecture=data.image.architecture,
            is_local=data.image.is_local,
            digest=data.image.digest,
            labels=[KVPair(key=kvpair.key, value=kvpair.value) for kvpair in data.image.labels],
            aliases=data.image.aliases,
            size_bytes=data.image.size_bytes,
            status=data.image.status,
            resource_limits=[
                ResourceLimit(
                    key=resource_limit.key, min=resource_limit.min, max=resource_limit.max
                )
                for resource_limit in data.image.resource_limits
            ],
            supported_accelerators=data.image.supported_accelerators,
            installed=data.agent_install_status.installed,
            installed_agents=data.agent_install_status.agent_names
            if data.agent_install_status.agent_names
            else None,
            # legacy
            hash=data.image.digest,
        )

    @classmethod
    async def batch_load_by_canonical(
        cls,
        graph_ctx: GraphQueryContext,
        image_names: Sequence[str],
        filter_by_statuses: list[ImageStatus] | None = None,
    ) -> list[Self]:
        if filter_by_statuses is None:
            filter_by_statuses = [ImageStatus.ALIVE]
        result = await graph_ctx.processors.image.get_images_by_canonicals.wait_for_complete(
            GetImagesByCanonicalsAction(
                image_canonicals=list(image_names),
                image_status=filter_by_statuses,
            )
        )
        return [
            cls.from_image_with_agent_install_status(img)
            for img in result.images_with_agent_install_status
        ]

    @classmethod
    async def batch_load_by_image_ref(
        cls,
        graph_ctx: GraphQueryContext,
        image_refs: Sequence[ImageRef],
        filter_by_statuses: list[ImageStatus] | None = None,
    ) -> Sequence[Image | None]:
        if filter_by_statuses is None:
            filter_by_statuses = [ImageStatus.ALIVE]
        image_names = [x.canonical for x in image_refs]
        return await cls.batch_load_by_canonical(graph_ctx, image_names, filter_by_statuses)

    @classmethod
    async def load_item_by_id(
        cls,
        ctx: GraphQueryContext,
        id: UUID,
        filter_by_statuses: list[ImageStatus] | None = None,
    ) -> Image:
        if filter_by_statuses is None:
            filter_by_statuses = [ImageStatus.ALIVE]
        result = await ctx.processors.image.get_image_by_id.wait_for_complete(
            GetImageByIdAction(
                image_id=ImageID(id),
                image_status=filter_by_statuses,
            )
        )
        return cls.from_image_with_agent_install_status(result.image_with_agent_install_status)

    @classmethod
    async def load_item(
        cls,
        ctx: GraphQueryContext,
        reference: str,
        architecture: str,
        filter_by_statuses: list[ImageStatus] | None = None,
    ) -> Image:
        if filter_by_statuses is None:
            filter_by_statuses = [ImageStatus.ALIVE]
        result = await ctx.processors.image.get_image_by_identifier.wait_for_complete(
            GetImageByIdentifierAction(
                image_identifier=ImageIdentifier(reference, architecture),
                image_status=filter_by_statuses,
            )
        )
        return cls.from_image_with_agent_install_status(result.image_with_agent_install_status)

    @classmethod
    async def load_all(
        cls,
        ctx: GraphQueryContext,
        *,
        types: set[ImageLoadFilter] | None = None,
        filter_by_statuses: list[ImageStatus] | None = None,
    ) -> Sequence[Image]:
        if filter_by_statuses is None:
            filter_by_statuses = [ImageStatus.ALIVE]
        if types is None:
            types = set()
        result = await ctx.processors.image.get_all_images.wait_for_complete(
            GetAllImagesAction(status_filter=filter_by_statuses)
        )
        all_items = [cls.from_image_with_agent_install_status(img) for img in result.data.values()]
        return [item for item in all_items if item.matches_filter(ctx, types)]

    @staticmethod
    async def filter_allowed(
        ctx: GraphQueryContext,
        items: Sequence[Image],
        domain_name: str,
    ) -> Sequence[Image]:
        from ai.backend.manager.models.domain import domains

        async with ctx.db.begin() as conn:
            query = (
                sa.select(domains.c.allowed_docker_registries)
                .select_from(domains)
                .where(domains.c.name == domain_name)
            )
            result = await conn.execute(query)
            allowed_docker_registries = result.scalar()

        if allowed_docker_registries is None:
            return []

        filtered_items: list[Image] = [
            item for item in items if item.registry in allowed_docker_registries
        ]

        return filtered_items

    def matches_filter(
        self,
        ctx: GraphQueryContext,
        load_filters: set[ImageLoadFilter],
    ) -> bool:
        """
        Determine if the image is filtered according to the `load_filters` parameter.
        """
        user_role = ctx.user["role"]

        # If the image filtered by any of its labels, return False early.
        # If the image is not filtered and is determiend to be valid by any of its labels, `is_valid = True`.
        is_valid = ImageLoadFilter.GENERAL in load_filters
        for label in self.labels:
            match label.key:
                case LabelName.FEATURES if KernelFeatures.OPERATION.value in label.value:
                    if ImageLoadFilter.OPERATIONAL in load_filters:
                        is_valid = True
                    else:
                        return False
                case LabelName.CUSTOMIZED_OWNER:
                    if (
                        ImageLoadFilter.CUSTOMIZED not in load_filters
                        and ImageLoadFilter.CUSTOMIZED_GLOBAL not in load_filters
                    ):
                        return False
                    if ImageLoadFilter.CUSTOMIZED in load_filters:
                        if label.value == f"user:{ctx.user['uuid']}":
                            is_valid = True
                        else:
                            return False
                    if ImageLoadFilter.CUSTOMIZED_GLOBAL in load_filters:
                        if user_role == UserRole.SUPERADMIN:
                            is_valid = True
                        else:
                            return False
        return is_valid

    @property
    def is_customized_image(self) -> bool:
        for label in self.labels:
            if label.key == LabelName.CUSTOMIZED_OWNER.value:
                return True
        return False


class ImagePermissionValueField(graphene.Scalar):  # type: ignore[misc]
    class Meta:
        description = f"Added in 25.3.0. One of {[val.value for val in ImagePermission]}."

    @staticmethod
    def serialize(val: ImagePermission) -> str:
        return val.value

    @staticmethod
    def parse_literal(
        node: Any, _variables: dict[str, Any] | None = None
    ) -> ImagePermission | None:
        if isinstance(node, graphql.language.ast.StringValueNode):
            return ImagePermission(node.value)
        return None

    @staticmethod
    def parse_value(value: str) -> ImagePermission:
        return ImagePermission(value)


@graphene_federation.key("id")
class ImageNode(graphene.ObjectType):  # type: ignore[misc]
    class Meta:
        interfaces = (AsyncNode,)

    row_id = graphene.UUID(description="Added in 24.03.4. The undecoded id value stored in DB.")
    name = graphene.String(deprecation_reason="Deprecated since 24.12.0. use `namespace` instead")
    namespace = graphene.String(description="Added in 24.12.0.")
    base_image_name = graphene.String(description="Added in 24.12.0.")
    project = graphene.String(description="Added in 24.03.10.")
    humanized_name = graphene.String()
    tag = graphene.String()
    tags = graphene.List(KVPair, description="Added in 24.12.0.")
    version = graphene.String(description="Added in 24.12.0.")
    registry = graphene.String()
    architecture = graphene.String()
    is_local = graphene.Boolean()
    digest = graphene.String()
    labels = graphene.List(KVPair)
    size_bytes = BigInt()
    status = graphene.String(description="Added in 25.4.0.")
    resource_limits = graphene.List(ResourceLimit)
    supported_accelerators = graphene.List(graphene.String)
    aliases = graphene.List(
        graphene.String, description="Added in 24.03.4. The array of image aliases."
    )

    permissions = graphene.List(
        ImagePermissionValueField,
        description=f"Added in 25.3.0. One of {[val.value for val in ImagePermission]}.",
    )
    installed = graphene.Boolean(
        description="Added in 25.11.0. Indicates if the image is installed on any Agent."
    )
    type = ImageTypeEnum(description="Added in 25.12.0.")

    @property
    def _canonical(self) -> str:
        image_ref = ImageRef(
            self.base_image_name,
            self.project,
            self.tag,
            self.registry,
            self.architecture,
            self.is_local,
        )
        return image_ref.canonical

    @classmethod
    async def _batch_load_installed_agents(
        cls, ctx: GraphQueryContext, image_ids: Sequence[ImageID]
    ) -> list[set[AgentId]]:
        result = await ctx.processors.image.get_image_installed_agents.wait_for_complete(
            GetImageInstalledAgentsAction(image_ids=list(image_ids))
        )
        installed_agent_ids_per_image: Mapping[ImageID, set[AgentId]] = result.data
        return [installed_agent_ids_per_image.get(image_id, set()) for image_id in image_ids]

    async def resolve_installed(self, info: graphene.ResolveInfo) -> bool:
        graph_ctx: GraphQueryContext = info.context
        loader = graph_ctx.dataloader_manager.get_loader_by_func(
            graph_ctx, self._batch_load_installed_agents
        )
        agent_ids = await loader.load(self.row_id)
        agent_ids = cast(set[AgentId] | None, agent_ids)
        return agent_ids is not None and len(agent_ids) > 0

    @classmethod
    async def batch_load_by_name_and_arch(
        cls,
        graph_ctx: GraphQueryContext,
        name_and_arch: Sequence[tuple[str, str]],
        filter_by_statuses: list[ImageStatus] | None = None,
    ) -> Sequence[Sequence[ImageNode]]:
        if filter_by_statuses is None:
            filter_by_statuses = [ImageStatus.ALIVE]
        query = (
            sa.select(ImageRow)
            .where(sa.tuple_(ImageRow.name, ImageRow.architecture).in_(name_and_arch))
            .options(selectinload(ImageRow.aliases))
        )
        if filter_by_statuses:
            query = query.where(ImageRow.status.in_(filter_by_statuses))

        async with graph_ctx.db.begin_readonly_session() as db_session:
            return await batch_multiresult_in_scalar_stream(
                graph_ctx,
                db_session,
                query,
                cls,
                name_and_arch,
                lambda row: (row.name, row.architecture),
            )

    @classmethod
    async def batch_load_by_image_identifier(
        cls,
        graph_ctx: GraphQueryContext,
        image_ids: Sequence[ImageIdentifier],
        filter_by_statuses: list[ImageStatus] | None = None,
    ) -> Sequence[Sequence[ImageNode]]:
        if filter_by_statuses is None:
            filter_by_statuses = [ImageStatus.ALIVE]
        name_and_arch_tuples = [(img.canonical, img.architecture) for img in image_ids]
        return await cls.batch_load_by_name_and_arch(
            graph_ctx, name_and_arch_tuples, filter_by_statuses
        )

    @overload
    @classmethod
    def from_row(cls, graph_ctx: GraphQueryContext, row: ImageRow) -> Self: ...

    @overload
    @classmethod
    def from_row(
        cls,
        graph_ctx: GraphQueryContext,
        row: ImageRow,
        *,
        permissions: Iterable[ImagePermission] | None = None,
    ) -> ImageNode: ...

    @overload
    @classmethod
    def from_row(
        cls,
        graph_ctx: GraphQueryContext,
        row: None,
        *,
        permissions: Iterable[ImagePermission] | None = None,
    ) -> None: ...

    @classmethod
    def from_row(
        cls,
        graph_ctx: GraphQueryContext,
        row: ImageRow | None,
        *,
        permissions: Iterable[ImagePermission] | None = None,
    ) -> ImageNode | None:
        if row is None:
            return None
        image_ref = row.image_ref
        version, ptag_set = image_ref.tag_set
        image_type = row.labels.get(LabelName.ROLE, ImageType.COMPUTE.value)

        return cls(
            id=row.id,
            row_id=row.id,
            name=row.image,
            namespace=row.image,
            base_image_name=image_ref.name,
            project=row.project,
            humanized_name=row.image,
            tag=row.tag,
            tags=[KVPair(key=k, value=v) for k, v in ptag_set.items()],
            version=version,
            registry=row.registry,
            architecture=row.architecture,
            is_local=row.is_local,
            digest=row.trimmed_digest or None,
            labels=[KVPair(key=k, value=v) for k, v in row.labels.items()],
            size_bytes=row.size_bytes,
            resource_limits=[
                ResourceLimit(key=k, min=v.get("min", Decimal(0)), max=Decimal("Infinity"))
                for k, v in row.resources.items()
            ],
            supported_accelerators=row.accelerators.split(",") if row.accelerators else ["*"],
            aliases=[alias_row.alias for alias_row in row.aliases],
            permissions=[] if permissions is None else permissions,
            status=row.status,
            type=image_type,
        )

    @classmethod
    def from_legacy_image(
        cls, image: Image, *, permissions: Iterable[ImagePermission] | None = None
    ) -> ImageNode:
        labels: dict[str, str] = {
            kvpair.key: kvpair.value for kvpair in cast(list[Any], image.labels)
        }
        image_type = labels.get(LabelName.ROLE, ImageType.COMPUTE.value)
        return cls(
            id=image.id,
            row_id=image.id,
            name=image.name,
            namespace=image.namespace,
            base_image_name=image.base_image_name,
            humanized_name=image.humanized_name,
            tag=image.tag,
            tags=image.tags,
            version=image.version,
            project=image.project,
            registry=image.registry,
            architecture=image.architecture,
            is_local=image.is_local,
            digest=image.digest,
            labels=image.labels,
            size_bytes=image.size_bytes,
            resource_limits=image.resource_limits,
            supported_accelerators=image.supported_accelerators,
            aliases=image.aliases,
            permissions=[] if permissions is None else permissions,
            status=image.status,
            type=image_type,
        )

    @classmethod
    async def get_node(
        cls,
        info: graphene.ResolveInfo,
        id: ResolvedGlobalID,
        scope_id: ScopeType,
        permission: ImagePermission = ImagePermission.READ_ATTRIBUTE,
    ) -> Self | None:
        graph_ctx: GraphQueryContext = info.context

        _, image_id = id
        async with graph_ctx.db.connect() as db_conn:
            user = graph_ctx.user
            client_ctx = ClientContext(
                graph_ctx.db, user["domain_name"], user["uuid"], user["role"]
            )
            permission_ctx = await get_permission_ctx(db_conn, client_ctx, scope_id, permission)
            cond = permission_ctx.query_condition
            if cond is None:
                return None

            query = (
                sa.select(ImageRow)
                .where(cond & (ImageRow.id == UUID(image_id)))
                .options(selectinload(ImageRow.aliases))
            )

            async with graph_ctx.db.begin_readonly_session() as db_session:
                image_row = await db_session.scalar(query)
                if image_row is None:
                    return None

                return cls.from_row(
                    graph_ctx,
                    image_row,
                    permissions=await permission_ctx.calculate_final_permission(image_row),
                )

    @classmethod
    async def get_connection(
        cls,
        info: graphene.ResolveInfo,
        scope_id: ScopeType,
        permission: ImagePermission,
        filter_by_statuses: list[ImageStatus] | None = None,
        filter_expr: str | None = None,
        order_expr: str | None = None,
        offset: int | None = None,
        after: str | None = None,
        first: int | None = None,
        before: str | None = None,
        last: int | None = None,
    ) -> ConnectionResolverResult[Self]:
        if filter_by_statuses is None:
            filter_by_statuses = [ImageStatus.ALIVE]
        graph_ctx: GraphQueryContext = info.context
        _filter_arg = (
            FilterExprArg(filter_expr, QueryFilterParser(_queryfilter_fieldspec))
            if filter_expr is not None
            else None
        )
        _order_expr = (
            OrderExprArg(order_expr, QueryOrderParser(_queryorder_colmap))
            if order_expr is not None
            else None
        )
        (
            query,
            cnt_query,
            _,
            cursor,
            pagination_order,
            page_size,
        ) = generate_sql_info_for_gql_connection(
            info,
            ImageRow,
            ImageRow.id,
            _filter_arg,
            _order_expr,
            offset,
            after=after,
            first=first,
            before=before,
            last=last,
        )
        async with graph_ctx.db.connect() as db_conn:
            user = graph_ctx.user
            client_ctx = ClientContext(
                graph_ctx.db, user["domain_name"], user["uuid"], user["role"]
            )
            permission_ctx = await get_permission_ctx(db_conn, client_ctx, scope_id, permission)
            cond = permission_ctx.query_condition
            if cond is None:
                return ConnectionResolverResult([], cursor, pagination_order, page_size, 0)
            query = query.where(cond).options(selectinload(ImageRow.aliases))
            cnt_query = cnt_query.where(cond)

            if filter_by_statuses:
                query = query.where(ImageRow.status.in_(filter_by_statuses))
                cnt_query = cnt_query.where(ImageRow.status.in_(filter_by_statuses))

            async with graph_ctx.db.begin_readonly_session(db_conn) as db_session:
                image_rows = (await db_session.scalars(query)).all()
                total_cnt = await db_session.scalar(cnt_query)
                result: list[Self] = [
                    cls.from_row(
                        graph_ctx,
                        row,
                        permissions=await permission_ctx.calculate_final_permission(row),
                    )
                    for row in image_rows
                ]
        return ConnectionResolverResult(result, cursor, pagination_order, page_size, total_cnt)

    # TODO: Introduce access control logic considering scope and permission
    async def __resolve_reference(self, info: graphene.ResolveInfo, **kwargs: Any) -> Image:
        ctx: GraphQueryContext = info.context
        _, image_id = AsyncNode.resolve_global_id(info, self.id)
        action_result = await ctx.processors.image.get_image_by_id.wait_for_complete(
            GetImageByIdAction(
                image_id=ImageID(UUID(image_id)),
                image_status=None,
            )
        )
        image_data = action_result.image_with_agent_install_status.image
        return ImageNode.from_row(ctx, ImageRow.from_dataclass_with_details(image_data))


class ImageConnection(Connection):
    class Meta:
        node = ImageNode
        description = "Added in 25.3.0."


class ForgetImageById(graphene.Mutation):  # type: ignore[misc]
    """Added in 24.03.0."""

    allowed_roles = (
        UserRole.SUPERADMIN,
        UserRole.ADMIN,
        UserRole.USER,
    )

    class Arguments:
        image_id = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    image = graphene.Field(ImageNode, description="Added since 24.03.1.")

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        image_id: str,
    ) -> ForgetImageById:
        log.info("forget image {0} by API request", image_id)
        image_uuid = extract_object_uuid(info, image_id, "image")

        ctx: GraphQueryContext = info.context

        result = await ctx.processors.image.forget_image_by_id.wait_for_complete(
            ForgetImageByIdAction(image_id=ImageID(image_uuid))
        )

        return ForgetImageById(
            ok=True, msg="", image=ImageNode.from_row(ctx, ImageRow.from_dataclass(result.image))
        )


class ForgetImage(graphene.Mutation):  # type: ignore[misc]
    """
    Deprecated since 25.4.0. Use `forget_image_by_id` instead.
    """

    allowed_roles = (
        UserRole.SUPERADMIN,
        UserRole.ADMIN,
        UserRole.USER,
    )

    class Arguments:
        reference = graphene.String(required=True)
        architecture = graphene.String(
            default_value=None,
            description="Changed to nullable in 26.1. If not provided, defaults to the Manager's architecture.",
        )

    ok = graphene.Boolean()
    msg = graphene.String()
    image = graphene.Field(ImageNode, description="Added since 24.03.1.")

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        reference: str,
        architecture: str | None,
    ) -> ForgetImage:
        log.info("forget image {0} by API request", reference)
        ctx: GraphQueryContext = info.context
        arch = architecture if architecture is not None else DEFAULT_IMAGE_ARCH

        result = await ctx.processors.image.forget_image.wait_for_complete(
            ForgetImageAction(
                reference=reference,
                architecture=arch,
            )
        )

        return ForgetImage(
            ok=True, msg="", image=ImageNode.from_row(ctx, ImageRow.from_dataclass(result.image))
        )


class PurgeImageOptions(graphene.InputObjectType):  # type: ignore[misc]
    """
    Added in 25.10.0.
    """

    remove_from_registry = graphene.Boolean(
        default_value=False,
        description="Untag the deleted image from the registry. Only available in the HarborV2 registry.",
    )


class PurgeImageById(graphene.Mutation):  # type: ignore[misc]
    """Added in 25.4.0."""

    allowed_roles = (
        UserRole.SUPERADMIN,
        UserRole.ADMIN,
        UserRole.USER,
    )

    class Arguments:
        image_id = graphene.String(required=True)
        options = PurgeImageOptions(
            required=False,
            default_value={"remove_from_registry": False},
            description="Added in 25.10.0.",
        )

    image = graphene.Field(ImageNode)

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        image_id: str,
        options: PurgeImageOptions,
    ) -> PurgeImageById:
        log.info("purge image row {0} by API request", image_id)
        image_uuid = extract_object_uuid(info, image_id, "image")

        ctx: GraphQueryContext = info.context
        result = await ctx.processors.image.purge_image_by_id.wait_for_complete(
            PurgeImageByIdAction(
                image_id=ImageID(image_uuid),
            )
        )

        if options.remove_from_registry:
            await ctx.processors.image.untag_image_from_registry.wait_for_complete(
                UntagImageFromRegistryAction(
                    image_id=ImageID(image_uuid),
                )
            )

        return PurgeImageById(image=ImageNode.from_row(ctx, ImageRow.from_dataclass(result.image)))


class UntagImageFromRegistry(graphene.Mutation):  # type: ignore[misc]
    """Deprecated since 25.10.0. Use `purge_image_by_id` with `remove_from_registry` option instead."""

    allowed_roles = (
        UserRole.SUPERADMIN,
        UserRole.ADMIN,
        UserRole.USER,
    )

    class Arguments:
        image_id = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    image = graphene.Field(ImageNode, description="Added since 24.03.1.")

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        image_id: str,
    ) -> UntagImageFromRegistry:
        image_uuid = extract_object_uuid(info, image_id, "image")

        log.info("remove image from registry {0} by API request", str(image_uuid))
        ctx: GraphQueryContext = info.context
        result = await ctx.processors.image.untag_image_from_registry.wait_for_complete(
            UntagImageFromRegistryAction(
                image_id=ImageID(image_uuid),
            )
        )

        return UntagImageFromRegistry(
            ok=True, msg="", image=ImageNode.from_row(ctx, ImageRow.from_dataclass(result.image))
        )


class PreloadImage(graphene.Mutation):  # type: ignore[misc]
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        references = graphene.List(graphene.String, required=True)
        target_agents = graphene.List(graphene.String, required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    task_id = graphene.String()

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        references: Sequence[str],
        target_agents: Sequence[str],
    ) -> PreloadImage:
        return PreloadImage(ok=False, msg="Not implemented.", task_id=None)


class UnloadImage(graphene.Mutation):  # type: ignore[misc]
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        references = graphene.List(graphene.String, required=True)
        target_agents = graphene.List(graphene.String, required=True)

    ok = graphene.Boolean()
    msg = graphene.String()
    task_id = graphene.String()

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        references: Sequence[str],
        target_agents: Sequence[str],
    ) -> UnloadImage:
        return UnloadImage(ok=False, msg="Not implemented.", task_id=None)


class RescanImages(graphene.Mutation):  # type: ignore[misc]
    allowed_roles = (UserRole.ADMIN, UserRole.SUPERADMIN)

    class Arguments:
        registry = graphene.String()
        project = graphene.String(description="Added in 25.1.0.")

    ok = graphene.Boolean()
    msg = graphene.String()
    task_id = graphene.UUID()

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        registry: str | None = None,
        project: str | None = None,
    ) -> RescanImages:
        log.info(
            "rescanning docker registry {0} by API request",
            f"(registry: {registry or 'all'}, project: {project or 'all'})",
        )
        ctx: GraphQueryContext = info.context

        async def _rescan_task(reporter: ProgressReporter) -> DispatchResult[Any]:
            if registry is None:
                all_registries = await ctx.processors.container_registry.load_all_container_registries.wait_for_complete(
                    LoadAllContainerRegistriesAction()
                )
                loaded_registries = all_registries.registries
            else:
                registries = await ctx.processors.container_registry.load_container_registries.wait_for_complete(
                    LoadContainerRegistriesAction(
                        registry=registry,
                        project=project,
                    )
                )
                loaded_registries = registries.registries

            rescanned_images: list[ImageData] = []
            errors: list[str] = []
            for registry_data in loaded_registries:
                action_result = (
                    await ctx.processors.container_registry.rescan_images.wait_for_complete(
                        RescanImagesAction(
                            registry=registry_data.registry_name,
                            project=registry_data.project,
                            progress_reporter=reporter,
                        )
                    )
                )
                for error in action_result.errors:
                    log.error(error)
                errors.extend(action_result.errors)
                rescanned_images.extend(action_result.images)

            rescanned_image_ids = [str(image.id) for image in rescanned_images]
            if errors:
                return DispatchResult.partial_success(rescanned_image_ids, errors)
            return DispatchResult.success(rescanned_image_ids)

        task_id = await ctx.background_task_manager.start(_rescan_task)
        return RescanImages(ok=True, msg="", task_id=task_id)


class AliasImage(graphene.Mutation):  # type: ignore[misc]
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        alias = graphene.String(required=True)
        target = graphene.String(required=True)
        architecture = graphene.String(
            default_value=None,
            description="Changed to nullable in 26.1. If not provided, defaults to the Manager's architecture.",
        )

    ok = graphene.Boolean()
    msg = graphene.String()

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        alias: str,
        target: str,
        architecture: str | None,
    ) -> AliasImage:
        log.info("alias image {0} -> {1} by API request", alias, target)
        ctx: GraphQueryContext = info.context
        arch = architecture if architecture is not None else DEFAULT_IMAGE_ARCH

        await ctx.processors.image.alias_image.wait_for_complete(
            AliasImageAction(
                image_canonical=target,
                architecture=arch,
                alias=alias,
            )
        )

        return AliasImage(ok=True, msg="")


class DealiasImage(graphene.Mutation):  # type: ignore[misc]
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        alias = graphene.String(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        alias: str,
    ) -> DealiasImage:
        log.info("dealias image {0} by API request", alias)
        ctx: GraphQueryContext = info.context

        await ctx.processors.image.dealias_image.wait_for_complete(
            DealiasImageAction(
                alias=alias,
            )
        )

        return DealiasImage(ok=True, msg="")


class ClearImages(graphene.Mutation):  # type: ignore[misc]
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        registry = graphene.String()

    ok = graphene.Boolean()
    msg = graphene.String()

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        registry: str,
    ) -> ClearImages:
        ctx: GraphQueryContext = info.context
        log.info("clear images from registry {0} by API request", registry)

        result = (
            await ctx.processors.container_registry.load_container_registries.wait_for_complete(
                LoadContainerRegistriesAction(
                    registry=registry,
                    project=None,
                )
            )
        )

        for registry_data in result.registries:
            await ctx.processors.container_registry.clear_images.wait_for_complete(
                ClearImagesAction(
                    registry=registry_data.registry_name,
                    project=registry_data.project,
                )
            )

        return ClearImages(ok=True, msg="")


class ModifyImageInput(graphene.InputObjectType):  # type: ignore[misc]
    name = graphene.String(required=False)
    registry = graphene.String(required=False)
    image = graphene.String(required=False)
    tag = graphene.String(required=False)
    architecture = graphene.String(required=False)
    is_local = graphene.Boolean(required=False)
    size_bytes = graphene.Int(required=False)
    type = graphene.String(required=False)
    digest = graphene.String(required=False)
    labels = graphene.List(lambda: KVPairInput, required=False)
    supported_accelerators = graphene.List(graphene.String, required=False)
    resource_limits = graphene.List(lambda: ResourceLimitInput, required=False)

    def to_updater_spec(self) -> ImageUpdaterSpec:
        resources_data: dict[str, Any] | UndefinedType = Undefined
        if self.resource_limits is not Undefined:
            resources_data = {}
            for limit_option in self.resource_limits:
                limit_data = {}
                if (
                    limit_option.min is not Undefined
                    and limit_option.min is not None
                    and len(limit_option.min) > 0
                ):
                    limit_data["min"] = limit_option.min
                if (
                    limit_option.max is not Undefined
                    and limit_option.max is not None
                    and len(limit_option.max) > 0
                ):
                    limit_data["max"] = limit_option.max
                resources_data[limit_option.key] = limit_data

        accelerators = (
            ",".join(self.supported_accelerators) if self.supported_accelerators else Undefined
        )
        labels = {label.key: label.value for label in self.labels} if self.labels else Undefined

        return ImageUpdaterSpec(
            name=OptionalState[str].from_graphql(self.name),
            registry=OptionalState[str].from_graphql(self.registry),
            image=OptionalState[str].from_graphql(self.image),
            tag=OptionalState[str].from_graphql(self.tag),
            architecture=OptionalState[str].from_graphql(self.architecture),
            is_local=OptionalState[bool].from_graphql(self.is_local),
            size_bytes=OptionalState[int].from_graphql(self.size_bytes),
            image_type=OptionalState[ImageType].from_graphql(self.type),
            config_digest=OptionalState[str].from_graphql(self.digest),
            labels=OptionalState[dict[str, Any]].from_graphql(labels),
            accelerators=TriState[str].from_graphql(
                accelerators,
            ),
            resources=OptionalState[dict[str, Any]].from_graphql(
                resources_data,
            ),
        )


class ModifyImage(graphene.Mutation):  # type: ignore[misc]
    allowed_roles = (UserRole.SUPERADMIN,)

    class Arguments:
        target = graphene.String(required=True, default_value=None)
        architecture = graphene.String(
            required=False,
            default_value=None,
            description="Changed to nullable in 26.1. If not provided, defaults to the Manager's architecture.",
        )
        props = ModifyImageInput(required=True)

    ok = graphene.Boolean()
    msg = graphene.String()

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        target: str,
        architecture: str | None,
        props: ModifyImageInput,
    ) -> AliasImage:
        ctx: GraphQueryContext = info.context
        log.info("modify image {0} by API request", target)
        arch = architecture if architecture is not None else DEFAULT_IMAGE_ARCH

        await ctx.processors.image.modify_image.wait_for_complete(
            ModifyImageAction(
                target=target,
                architecture=arch,
                updater_spec=props.to_updater_spec(),
            )
        )

        return ModifyImage(ok=True, msg="")


class PurgeImagesKey(graphene.InputObjectType):  # type: ignore[misc]
    """
    Added in 25.6.0.
    """

    agent_id = graphene.String(required=True)
    images = graphene.List(ImageRefType, required=True)


class PurgeImagesOptions(graphene.InputObjectType):  # type: ignore[misc]
    """
    Added in 25.6.0.
    """

    force = graphene.Boolean(
        default_value=False,
        description="Remove the images even if it is being used by stopped containers or has other tags, Added in 25.6.0.",
    )
    noprune = graphene.Boolean(
        default_value=False, description="Don't delete untagged parent images, Added in 25.6.0."
    )


class PurgeImagesPayload(graphene.ObjectType):  # type: ignore[misc]
    """
    Added in 25.6.0.
    """

    task_id = graphene.String()
    allowed_roles = (UserRole.SUPERADMIN, UserRole.ADMIN)


class PurgeImages(graphene.Mutation):  # type: ignore[misc]
    """
    Added in 25.4.0.
    """

    class Arguments:
        keys = graphene.List(PurgeImagesKey, required=True)
        options = PurgeImagesOptions(default_value={"force": False, "noprune": False})

    Output = PurgeImagesPayload

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        keys: list[PurgeImagesKey],
        options: PurgeImagesOptions,
    ) -> PurgeImagesPayload:
        ctx: GraphQueryContext = info.context
        agent_images = ", ".join(
            f"{key.agent_id}: [{', '.join(img.name for img in key.images)}]" for key in keys
        )

        log.info(f"purge images ({agent_images}) by API request")

        # Convert GraphQL input types to bgtask manifest types
        manifest_keys = [
            PurgeAgentSpec(
                agent_id=AgentId(key.agent_id),
                images=[
                    PurgeImageSpec(
                        name=img.name,
                        registry=img.registry or "",
                        architecture=img.architecture or "",
                    )
                    for img in key.images
                ],
            )
            for key in keys
        ]

        manifest = PurgeImagesManifest(
            keys=manifest_keys,
            force=options.force,
            noprune=options.noprune,
        )
        task_id = await ctx.background_task_manager.start_retriable(
            ManagerBgtaskName.PURGE_IMAGES,
            manifest,
        )
        return PurgeImagesPayload(task_id=task_id)


class ClearImageCustomResourceLimitKey(graphene.InputObjectType):  # type: ignore[misc]
    """
    Added in 25.6.0.
    """

    image_canonical = graphene.String(required=True)
    architecture = graphene.String(
        required=False,
        default_value=None,
        description="Changed to nullable in 26.1. If not provided, defaults to the Manager's architecture.",
    )


class ClearImageCustomResourceLimitPayload(graphene.ObjectType):  # type: ignore[misc]
    """
    Added in 25.6.0.
    """

    image_node = graphene.Field(ImageNode)
    allowed_roles = (UserRole.SUPERADMIN,)


class ClearImageCustomResourceLimit(graphene.Mutation):  # type: ignore[misc]
    """
    Added in 25.6.0.
    """

    class Arguments:
        key = ClearImageCustomResourceLimitKey(required=True)

    Output = ClearImageCustomResourceLimitPayload

    @staticmethod
    async def mutate(
        root: Any,
        info: graphene.ResolveInfo,
        key: ClearImageCustomResourceLimitKey,
    ) -> ClearImageCustomResourceLimitPayload:
        arch = key.architecture if key.architecture is not None else DEFAULT_IMAGE_ARCH
        log.info(
            f'clear custom resource limits for image "{key.image_canonical}" ({arch}) by API request',
        )
        ctx: GraphQueryContext = info.context
        result = await ctx.processors.image.clear_image_custom_resource_limit.wait_for_complete(
            ClearImageCustomResourceLimitAction(
                image_canonical=key.image_canonical,
                architecture=arch,
            )
        )
        return ClearImageCustomResourceLimitPayload(
            image_node=ImageNode.from_row(ctx, ImageRow.from_dataclass(result.image_data))
        )



from dataclasses import dataclass
from typing import override

from ai.backend.common.types import ImageID
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.image.types import ImageAliasData
from ai.backend.manager.services.image.actions.alias_base import ImageAliasAction


@dataclass
class DealiasImageAction(ImageAliasAction):
    alias: str

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.DELETE


@dataclass
class DealiasImageActionResult(BaseActionResult):
    image_id: ImageID
    image_alias: ImageAliasData

    @override
    def entity_id(self) -> str | None:
        return str(self.image_id)
